from probeye._setup_cfg import version_constraint_free_packages


def test_version_removal():
    p_in_list = [
        "numpy<2",
        "scipy!=1.10, >1.0",
        "pandas",
        "torch[dev]",
        "matplotlib[dev]<=2",
        "tensorflow==1.23rc",
    ]
    expected_p_out_list = [
        "numpy",
        "scipy",
        "pandas",
        "torch[dev]",
        "matplotlib[dev]",
        "tensorflow",
    ]
    for ii in range(len(p_in_list)):
        p_in_list_ii = p_in_list[: (ii + 1)]
        expected_p_out_list_ii = expected_p_out_list[: (ii + 1)]

        # .......................................................
        # Pure dangling list
        # .......................................................
        p_in = "\n" + "\n".join(p_in_list_ii)
        p_out_list_ii = version_constraint_free_packages(setup_cfg_packages=p_in)
        assert p_out_list_ii == expected_p_out_list_ii

        # with extra space
        p_in = "\n" + "\n ".join(p_in_list_ii)
        p_out_list_ii = version_constraint_free_packages(setup_cfg_packages=p_in)
        assert p_out_list_ii == expected_p_out_list_ii

        # .......................................................
        # Pure semi-colon separated list
        # .......................................................
        p_in = ";".join(p_in_list_ii)
        p_out_list_ii = version_constraint_free_packages(setup_cfg_packages=p_in)
        assert p_out_list_ii == expected_p_out_list_ii

        # .......................................................
        # Dangling list combined with semi-colon separated list
        # .......................................................
        if ii > 2:
            p_in = (
                "\n" + "\n".join(p_in_list_ii[:2]) + "\n" + ";".join(p_in_list_ii[2:])
            )
            p_out_list_ii = version_constraint_free_packages(setup_cfg_packages=p_in)
            assert p_out_list_ii == expected_p_out_list_ii

            # reversed order
            p_in = "\n".join(p_in_list_ii[:2]) + "\n" + ";".join(p_in_list_ii[2:])
            p_out_list_ii = version_constraint_free_packages(setup_cfg_packages=p_in)
            assert p_out_list_ii == expected_p_out_list_ii



# standard library imports
from loguru import logger
import unittest
import os

# third party imports
import numpy as np

# local imports
from probeye.definition.inference_problem import InferenceProblem
from probeye.subroutines import len_or_one
from probeye.subroutines import make_list
from probeye.subroutines import underlined_string
from probeye.subroutines import titled_table
from probeye.subroutines import replace_string_chars
from probeye.subroutines import simplified_list_string
from probeye.subroutines import simplified_dict_string
from probeye.subroutines import unvectorize_dict_values
from probeye.subroutines import sub_when_empty
from probeye.subroutines import dict2list
from probeye.subroutines import list2dict
from probeye.subroutines import pretty_time_delta
from probeye.subroutines import flatten
from probeye.subroutines import process_spatial_coordinates
from probeye.subroutines import translate_prms_def
from probeye.subroutines import print_probeye_header
from probeye.subroutines import logging_setup
from probeye.subroutines import print_dict_in_rows
from probeye.subroutines import add_index_to_tex_prm_name
from probeye.subroutines import check_for_uninformative_priors


class TestProblem(unittest.TestCase):
    def test_len_or_one(self):
        # check for objects with no len-method
        self.assertEqual(len_or_one(1), 1)
        self.assertEqual(len_or_one(1.0), 1)
        self.assertEqual(len_or_one(True), 1)
        self.assertEqual(len_or_one("a"), 1)
        self.assertEqual(len_or_one(None), 1)
        # check for lists with zero or more elements
        self.assertEqual(len_or_one([]), 0)
        self.assertEqual(len_or_one([1]), 1)
        self.assertEqual(len_or_one([1, 2]), 2)
        self.assertEqual(len_or_one([1, 2, 3]), 3)
        # check for the weird numpy case (the object np.array(1.0) does have len-method,
        # but when you do len(np.array(1.0)) you get an error)
        self.assertEqual(len_or_one(np.array(1.0)), 1)

    def test_make_list(self):
        # check main use for single non-list input
        self.assertEqual(make_list("a"), ["a"])
        # check common use for lists
        self.assertEqual(make_list(["a"]), ["a"])
        self.assertEqual(make_list(["a", "b"]), ["a", "b"])
        self.assertEqual(make_list([]), [])
        # check common use for tuples
        self.assertEqual(make_list(("a", "b")), ("a", "b"))
        self.assertEqual(make_list(()), ())

    def test_underlined_string(self):
        # check common use cases
        self.assertEqual(underlined_string("Test"), "\nTest\n====\n")
        self.assertEqual(
            underlined_string("ABC", n_empty_start=0, n_empty_end=2, symbol="-"),
            "ABC\n---\n\n",
        )

    def test_titled_table(self):
        # check common use case
        titled_table_str = titled_table(
            "Title", "Col Col", symbol="=", n_empty_start=2, n_empty_end=1
        )
        self.assertEqual(titled_table_str, "\n\nTitle\n=======\nCol Col\n\n")

    def test_replace_string_chars(self):
        # check common use case
        self.assertEqual(
            replace_string_chars("Test1233", replace={"T": "t", "3": "4"}), "test1244"
        )
        self.assertEqual(replace_string_chars("Test1233", remove=["1", "e"]), "Tst233")
        self.assertEqual(
            replace_string_chars("Test1233", replace={"3": "4"}, remove=["1"]),
            "Test244",
        )

    def test_simplified_list_string(self):
        # check common use case
        self.assertEqual(simplified_list_string(["a", "b"]), "a, b")
        self.assertEqual(simplified_list_string(["a"]), "a")

    def test_simplified_dict_string(self):
        # check common use case
        self.assertEqual(simplified_dict_string({"a": 1, "b": 2}), "a=1, b=2")
        self.assertEqual(simplified_dict_string({"a": 1.0}), "a=1.0")

    def test_unvectorize_dict_values(self):
        # check common use case
        self.assertEqual(
            unvectorize_dict_values({"x": [1, 2], "y": [3, 4]}),
            [{"x": 1, "y": 3}, {"x": 2, "y": 4}],
        )
        # check for cases with invalid input
        with self.assertRaises(RuntimeError):
            # values cannot have different lengths
            unvectorize_dict_values({"x": [1, 2, 3], "y": [3, 4]})

    def test_sub_when_empty(self):
        # check common use cases
        self.assertEqual(sub_when_empty("test"), "test")
        self.assertEqual(sub_when_empty("", empty_str="-"), "-")
        # check for cases with invalid input
        with self.assertRaises(TypeError):
            # noinspection PyTypeChecker
            sub_when_empty(1.0)

    def test_dict2list(self):
        # check common use cases
        self.assertEqual(dict2list({"a": 1, "b": 2}), [{"a": 1}, {"b": 2}])
        self.assertEqual(dict2list({"a": 1}), [{"a": 1}])
        # check for cases with invalid input
        with self.assertRaises(TypeError):
            # noinspection PyTypeChecker
            dict2list([{"a": 1}, {"b": 2}])

    def test_list2dict(self):
        # check common use cases
        self.assertEqual(list2dict(["a", "b"]), {"a": "a", "b": "b"})
        self.assertEqual(list2dict([{"a": "m"}, "b"]), {"a": "m", "b": "b"})
        self.assertEqual(list2dict([{"a": "m"}, {"b": "c"}]), {"a": "m", "b": "c"})
        self.assertEqual(list2dict({"a": "m", "b": "c"}), {"a": "m", "b": "c"})
        # each dict in the list must not contain more than one element
        with self.assertRaises(ValueError):
            list2dict([{"a": "m1", "b": "b"}])
        # the input cannot contain numbers
        with self.assertRaises(TypeError):
            list2dict(["a", 1.0])
        # the input must be of type list or dict
        with self.assertRaises(TypeError):
            # noinspection PyTypeChecker
            list2dict((1, 2))

    def test_pretty_time_delta(self):
        # check the different output formats
        self.assertEqual(pretty_time_delta(1), "1s")
        self.assertEqual(pretty_time_delta(62), "1m2s")
        self.assertEqual(pretty_time_delta(3723), "1h2m3s")
        self.assertEqual(pretty_time_delta(93780), "1d2h3m")

    def test_flatten(self):
        # test for a simple nested list
        nested_list_1 = [1, 2, [3, 4]]
        computed_result = flatten(nested_list_1)
        expected_result = [1, 2, 3, 4]
        self.assertEqual(computed_result, expected_result)
        # test for another simple nested list
        nested_list_2 = [[1, 2, [3, 4]]]
        computed_result = flatten(nested_list_2)
        expected_result = [1, 2, 3, 4]
        self.assertEqual(computed_result, expected_result)
        # test for a flat list
        flat_list = [1, 2, 3, 4]
        computed_result = flatten(flat_list)
        expected_result = [1, 2, 3, 4]
        self.assertEqual(computed_result, expected_result)
        # test for a simple nested numpy-array
        nested_array_1 = np.array([[1, 2], [3, 4]])
        computed_result = flatten(nested_array_1)
        expected_result = [1, 2, 3, 4]
        self.assertEqual(computed_result, expected_result)
        # test for another simple nested numpy-array
        nested_array_2 = np.array([[1, 2, 3, 4]])
        computed_result = flatten(nested_array_2)
        expected_result = [1, 2, 3, 4]
        self.assertEqual(computed_result, expected_result)
        # test for a flat numpy-array
        flat_array = np.array([1, 2, 3, 4])
        computed_result = flatten(flat_array)
        expected_result = [1, 2, 3, 4]
        self.assertEqual(computed_result, expected_result)
        # test for use with scalar
        scalar = 1.2
        computed_result = flatten(scalar)
        expected_result = [1.2]
        self.assertEqual(computed_result, expected_result)
        # test for the None-input
        none_arg = None
        computed_result = flatten(none_arg)
        expected_result = None
        self.assertEqual(computed_result, expected_result)
        # check invalid input (here: invalid type)
        with self.assertRaises(TypeError):
            flatten((1, 2, 3))

    def test_process_spatial_coordinates(self):
        # check for complete None-input
        coords_computed, order_computed = process_spatial_coordinates(
            x=None, y=None, z=None, coords=None
        )
        coords_expected = np.array([])
        order_expected = []
        self.assertTrue(
            np.allclose(coords_computed, coords_expected)
            and coords_computed.shape == coords_expected.shape
        )
        self.assertEqual(order_computed, order_expected)
        # check for only x being given as numpy array
        coords_computed, order_computed = process_spatial_coordinates(
            x=np.array([1, 2, 3]), y=None, z=None, coords=None
        )
        coords_expected = np.array([[1, 2, 3]])
        order_expected = ["x"]
        self.assertTrue(
            np.allclose(coords_computed, coords_expected)
            and coords_computed.shape == coords_expected.shape
        )
        self.assertEqual(order_computed, order_expected)
        # check for only x being given as list
        coords_computed, order_computed = process_spatial_coordinates(
            x=[1, 2, 3], y=None, z=None, coords=None
        )
        coords_expected = np.array([[1, 2, 3]])
        order_expected = ["x"]
        self.assertTrue(
            np.allclose(coords_computed, coords_expected)
            and coords_computed.shape == coords_expected.shape
        )
        self.assertEqual(order_computed, order_expected)
        # check for only y being given as numpy array
        coords_computed, order_computed = process_spatial_coordinates(
            x=None, y=np.array([1, 2, 3]), z=None, coords=None
        )
        coords_expected = np.array([[1, 2, 3]])
        order_expected = ["y"]
        self.assertTrue(
            np.allclose(coords_computed, coords_expected)
            and coords_computed.shape == coords_expected.shape
        )
        self.assertEqual(order_computed, order_expected)
        # check for only y being given as list
        coords_computed, order_computed = process_spatial_coordinates(
            x=None, y=[1, 2, 3], z=None, coords=None
        )
        coords_expected = np.array([[1, 2, 3]])
        order_expected = ["y"]
        self.assertTrue(
            np.allclose(coords_computed, coords_expected)
            and coords_computed.shape == coords_expected.shape
        )
        self.assertEqual(order_computed, order_expected)
        # check for only z being given as numpy array
        coords_computed, order_computed = process_spatial_coordinates(
            x=None, y=None, z=np.array([1, 2, 3]), coords=None
        )
        coords_expected = np.array([[1, 2, 3]])
        order_expected = ["z"]
        self.assertTrue(
            np.allclose(coords_computed, coords_expected)
            and coords_computed.shape == coords_expected.shape
        )
        self.assertEqual(order_computed, order_expected)
        # check for only z being given as list
        coords_computed, order_computed = process_spatial_coordinates(
            x=None, y=None, z=[1, 2, 3], coords=None
        )
        coords_expected = np.array([[1, 2, 3]])
        order_expected = ["z"]
        self.assertTrue(
            np.allclose(coords_computed, coords_expected)
            and coords_computed.shape == coords_expected.shape
        )
        self.assertEqual(order_computed, order_expected)
        # check for x and y being given as numpy arrays
        coords_computed, order_computed = process_spatial_coordinates(
            x=np.array([1, 2, 3]), y=np.array([4, 5, 6]), z=None, coords=None
        )
        coords_expected = np.array([[1, 2, 3], [4, 5, 6]])
        order_expected = ["x", "y"]
        self.assertTrue(
            np.allclose(coords_computed, coords_expected)
            and coords_computed.shape == coords_expected.shape
        )
        self.assertEqual(order_computed, order_expected)
        # check for x and y being given as lists
        coords_computed, order_computed = process_spatial_coordinates(
            x=[1, 2, 3], y=[4, 5, 6], z=None, coords=None
        )
        coords_expected = np.array([[1, 2, 3], [4, 5, 6]])
        order_expected = ["x", "y"]
        self.assertTrue(
            np.allclose(coords_computed, coords_expected)
            and coords_computed.shape == coords_expected.shape
        )
        self.assertEqual(order_computed, order_expected)
        # check for x and z being given as numpy arrays
        coords_computed, order_computed = process_spatial_coordinates(
            x=np.array([1, 2, 3]), y=None, z=np.array([4, 5, 6]), coords=None
        )
        coords_expected = np.array([[1, 2, 3], [4, 5, 6]])
        order_expected = ["x", "z"]
        self.assertTrue(
            np.allclose(coords_computed, coords_expected)
            and coords_computed.shape == coords_expected.shape
        )
        self.assertEqual(order_computed, order_expected)
        # check for x and z being given as lists
        coords_computed, order_computed = process_spatial_coordinates(
            x=[1, 2, 3], y=None, z=[4, 5, 6], coords=None
        )
        coords_expected = np.array([[1, 2, 3], [4, 5, 6]])
        order_expected = ["x", "z"]
        self.assertTrue(
            np.allclose(coords_computed, coords_expected)
            and coords_computed.shape == coords_expected.shape
        )
        self.assertEqual(order_computed, order_expected)
        # check for y and z being given as numpy arrays
        coords_computed, order_computed = process_spatial_coordinates(
            x=None, y=np.array([1, 2, 3]), z=np.array([4, 5, 6]), coords=None
        )
        coords_expected = np.array([[1, 2, 3], [4, 5, 6]])
        order_expected = ["y", "z"]
        self.assertTrue(
            np.allclose(coords_computed, coords_expected)
            and coords_computed.shape == coords_expected.shape
        )
        self.assertEqual(order_computed, order_expected)
        # check for y and z being given as lists
        coords_computed, order_computed = process_spatial_coordinates(
            x=None, y=[1, 2, 3], z=[4, 5, 6], coords=None
        )
        coords_expected = np.array([[1, 2, 3], [4, 5, 6]])
        order_expected = ["y", "z"]
        self.assertTrue(
            np.allclose(coords_computed, coords_expected)
            and coords_computed.shape == coords_expected.shape
        )
        self.assertEqual(order_computed, order_expected)
        # check for x, y and z being given as numpy arrays
        coords_computed, order_computed = process_spatial_coordinates(
            x=np.array([1, 2, 3]),
            y=np.array([4, 5, 6]),
            z=np.array([7, 8, 9]),
            coords=None,
        )
        coords_expected = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
        order_expected = ["x", "y", "z"]
        self.assertTrue(
            np.allclose(coords_computed, coords_expected)
            and coords_computed.shape == coords_expected.shape
        )
        self.assertEqual(order_computed, order_expected)
        # check for y and z being given as lists
        coords_computed, order_computed = process_spatial_coordinates(
            x=[1, 2, 3], y=[4, 5, 6], z=[7, 8, 9], coords=None
        )
        coords_expected = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
        order_expected = ["x", "y", "z"]
        self.assertTrue(
            np.allclose(coords_computed, coords_expected)
            and coords_computed.shape == coords_expected.shape
        )
        self.assertEqual(order_computed, order_expected)
        # check for coords being given as numpy array with new order
        coords_computed, order_computed = process_spatial_coordinates(
            coords=np.array([[4, 5, 6], [1, 2, 3], [7, 8, 9]]), order=("y", "x", "z")
        )
        coords_expected = np.array([[4, 5, 6], [1, 2, 3], [7, 8, 9]])
        order_expected = ["y", "x", "z"]
        self.assertTrue(
            np.allclose(coords_computed, coords_expected)
            and coords_computed.shape == coords_expected.shape
        )
        self.assertEqual(order_computed, order_expected)
        # check invalid input (here: invalid type)
        with self.assertRaises(TypeError):
            process_spatial_coordinates(coords=(1, 2, 3, 4, 5))
        # check invalid input (here: inconsistent individual lengths)
        with self.assertRaises(RuntimeError):
            process_spatial_coordinates(x=[1], y=[1, 2])

    def test_translate_prms_def(self):
        # valid use case: single string
        prms_def_given = "sigma"
        expected_result = {"sigma": "sigma"}
        computed_result, _ = translate_prms_def(prms_def_given)
        self.assertEqual(computed_result, expected_result)
        # valid use case: single string as list
        prms_def_given = ["sigma"]
        expected_result = {"sigma": "sigma"}
        computed_result, _ = translate_prms_def(prms_def_given)
        self.assertEqual(computed_result, expected_result)
        # valid use case: single 1-element dict
        prms_def_given = {"sigma": "sigma"}
        expected_result = {"sigma": "sigma"}
        computed_result, _ = translate_prms_def(prms_def_given)
        self.assertEqual(computed_result, expected_result)
        # valid use case: multiple-element dict
        prms_def_given = {"sigma": "std", "mu": "mean"}
        expected_result = {"sigma": "std", "mu": "mean"}
        computed_result, _ = translate_prms_def(prms_def_given)
        self.assertEqual(computed_result, expected_result)
        # valid use case: multiple-element list of strings
        prms_def_given = ["sigma", "mu"]
        expected_result = {"sigma": "sigma", "mu": "mu"}
        computed_result, _ = translate_prms_def(prms_def_given)
        self.assertEqual(computed_result, expected_result)
        # valid use case: mixed multiple-element list
        prms_def_given = ["sigma", {"mu": "mean"}]
        expected_result = {"sigma": "sigma", "mu": "mean"}
        computed_result, _ = translate_prms_def(prms_def_given)
        self.assertEqual(computed_result, expected_result)
        # invalid use case: list with 2-element dict
        with self.assertRaises(ValueError):
            translate_prms_def([{"sigma": "sigma", "mu": "mean"}])

    def test_print_probeye_header(self):
        # simply check that the defaults work
        print_probeye_header(use_logger=True)
        print_probeye_header(use_logger=False)

    def test_logging_setup(self):
        # check that the log-file option works
        log_dir = os.path.dirname(__file__)
        log_file = os.path.join(log_dir, "logfile.txt")
        logging_setup(log_file=log_file)
        logger.info("This text should also be written to the log-file.")
        self.assertTrue(os.path.exists(log_file))
        # now, check the overwrite-option
        logger.stop()  # necessary to remove the logfile
        logging_setup(log_file=log_file, overwrite_log_file=True)
        logger.info("This text should be overwriting the previous one.")
        self.assertTrue(os.path.exists(log_file))
        logger.stop()  # necessary to remove the logfile
        os.remove(log_file)

    def test_print_dict_in_rows(self):
        # this check covers the val_fmt options
        print_dict_in_rows({"a": 1.2345, "b": 3.45365}, val_fmt=".2f")

    def test_add_index_to_tex_prm_name(self):
        # check normal use case 1
        computed_result = add_index_to_tex_prm_name(r"$\alpha$", 1)
        expected_result = r"$\alpha_1$"
        self.assertEqual(computed_result, expected_result)
        # check normal use case 2
        computed_result = add_index_to_tex_prm_name("$\\alpha$", 2)
        expected_result = "$\\alpha_2$"
        self.assertEqual(computed_result, expected_result)
        # check normal use case 3
        computed_result = add_index_to_tex_prm_name("$a$", 3)
        expected_result = "$a_3$"
        self.assertEqual(computed_result, expected_result)
        # check normal use case 4 (note the missing '$'-signs)
        computed_result = add_index_to_tex_prm_name("a", 4)
        expected_result = "a (4)"
        self.assertEqual(computed_result, expected_result)
        # check normal use case 5
        computed_result = add_index_to_tex_prm_name("$a_c$", 5)
        expected_result = "$a_c$ (5)"
        self.assertEqual(computed_result, expected_result)
        # check normal use case 6
        computed_result = add_index_to_tex_prm_name("$\\gamma^c$", 6)
        expected_result = "$\\gamma^c$ (6)"
        self.assertEqual(computed_result, expected_result)

    def test_check_for_uninformative_priors(self):
        # check if the detection works as expected
        problem = InferenceProblem("Problem with uninformative prior")
        problem.add_parameter(
            "sigma", "noise", prior=("uniform", {"low": 0.1, "high": 0.8})
        )
        check_for_uninformative_priors(problem)  # no error should be raised
        problem.add_parameter("m", "model")  # uninformative prior
        with self.assertRaises(RuntimeError):
            check_for_uninformative_priors(problem)


if __name__ == "__main__":
    unittest.main()



# standard library
import logging
import unittest

# third party imports
import numpy as np

# local imports
from probeye.definition.forward_model import ForwardModelBase
from probeye.definition.sensor import Sensor
from probeye.definition.inference_problem import InferenceProblem
from probeye.definition.noise_model import NormalNoiseModel
from probeye.inference.scipy_.solver import ScipySolver


class TestProblem(unittest.TestCase):
    def test_scipy_solver(self):

        # define a simple linear model for this case
        class LinearModel(ForwardModelBase):
            def response(self, inp):
                return {"y": inp["m"] * inp["x"] + inp["b"]}

        # define parameters with with an uninformative prior
        problem = InferenceProblem("Problem with uninformative prior")
        problem.add_parameter("m", "model")  # uninformative prior
        problem.add_parameter(
            "b", "model", prior=("normal", {"loc": 1.0, "scale": 1.0})
        )
        problem.add_parameter(
            "sigma", "noise", prior=("uniform", {"low": 0.1, "high": 0.8})
        )

        # add forward model and noise model
        isensor, osensor = Sensor("x"), Sensor("y")
        linear_model = LinearModel(["m", "b"], [isensor], [osensor])
        problem.add_forward_model("LinearModel", linear_model)
        problem.add_noise_model(
            NormalNoiseModel(prms_def={"sigma": "std"}, sensors=osensor)
        )

        # add experimental data
        np.random.seed(1)
        x_test = np.linspace(0.0, 1.0, 10)
        y_true = linear_model({isensor.name: x_test, "m": 2.5, "b": 1.7})[osensor.name]
        y_test = np.random.normal(loc=y_true, scale=0.5)
        problem.add_experiment(
            f"TestSeries_1",
            fwd_model_name="LinearModel",
            sensor_values={isensor.name: x_test, osensor.name: y_test},
        )

        # test the get_start_values method for given x0_dict
        scipy_solver = ScipySolver(problem)
        x0, x0_dict = scipy_solver.get_start_values(
            x0_dict={"m": 2.5, "b": 1.5, "sigma": 0.6}
        )
        self.assertTrue(np.allclose(x0, np.array([2.5, 1.5, 0.6])))
        self.assertEqual(x0_dict, {"m": 2.5, "b": 1.5, "sigma": 0.6})

        # test the get_start_values method for automatic derivation
        scipy_solver = ScipySolver(problem)
        x0, x0_dict = scipy_solver.get_start_values(x0_prior="mean")
        self.assertTrue(np.allclose(x0, np.array([1.0, 1.0, 0.45])))
        expected_x0_dict = {"m": 1.0, "b": 1.0, "sigma": 0.45}
        for key, val in x0_dict.items():
            self.assertAlmostEqual(val, expected_x0_dict[key])

        # check that the solver can be run now; note that the additional solver option
        # is a default value, and only provided to check that this argument-pipeline
        # works
        scipy_solver.run_max_likelihood(solver_options={"jac": None})

        # check the warning when the problem cannot be deep-copied
        problem.no_deepcopy_possible = (i for i in (1, 2, 3))
        ScipySolver(problem)  # this should result in a warning

    def test_evaluate_model_response(self):
        # prepare for checks
        p = InferenceProblem("TestProblem")
        p.add_parameter("a0", "model", prior=("normal", {"loc": 0, "scale": 1}))
        p.add_parameter("a1", "model", prior=("normal", {"loc": 0, "scale": 1}))
        p.add_parameter("a2", "model", prior=("normal", {"loc": 0, "scale": 1}))
        p.add_parameter("sigma", "noise", const=1.0)

        class FwdModel(ForwardModelBase):
            def response(self, inp):
                x = inp["x"]
                a0 = inp["a0"]
                a1 = inp["a1"]
                a2 = inp["a2"]
                return {"y": a0 + a1 * x + a2 * x ** 2}

        # add forward and noise model
        fwd_model = FwdModel(["a0", "a1", "a2"], Sensor("x"), Sensor("y"))
        p.add_forward_model("FwdModel", fwd_model)
        p.add_noise_model(
            NormalNoiseModel(prms_def={"sigma": "std"}, sensors=Sensor("y"))
        )

        # add experiment_names
        p.add_experiment(
            "Exp1", sensor_values={"x": 1, "y": 2}, fwd_model_name="FwdModel"
        )
        p.add_experiment(
            "Exp2", sensor_values={"x": 2, "y": 3}, fwd_model_name="FwdModel"
        )
        p.add_experiment(
            "Exp3", sensor_values={"x": [1, 2], "y": [1, 2]}, fwd_model_name="FwdModel"
        )

        # initialize the solver object
        pyro_solver = ScipySolver(p)

        # perform a check for all experiments
        a0_value, a1_value, a2_value = 1, 2, 3
        theta = np.array([a0_value, a1_value, a2_value])
        computed_result = pyro_solver.evaluate_model_response(theta)
        expected_result = {
            "Exp1": {"y": 6},
            "Exp2": {"y": 17},
            "Exp3": {"y": np.array([6, 17])},
        }

        # check for each item, because the last contains an array, which results in an
        # error is you do assertEqual on the whole thing
        self.assertEqual(computed_result["Exp1"], expected_result["Exp1"])
        self.assertEqual(computed_result["Exp2"], expected_result["Exp2"])
        self.assertEqual(
            computed_result["Exp3"]["y"][0], expected_result["Exp3"]["y"][0]
        )
        self.assertEqual(
            computed_result["Exp3"]["y"][1], expected_result["Exp3"]["y"][1]
        )

        # perform a check for a subset of experiments
        computed_result = pyro_solver.evaluate_model_response(
            theta, experiment_names="Exp3"
        )
        expected_result = {"Exp3": {"y": np.array([6, 17])}}
        self.assertEqual(
            computed_result["Exp3"]["y"][0], expected_result["Exp3"]["y"][0]
        )
        self.assertEqual(
            computed_result["Exp3"]["y"][1], expected_result["Exp3"]["y"][1]
        )


if __name__ == "__main__":
    unittest.main()



# standard library imports
from typing import Union, List, Optional, TYPE_CHECKING

# third party imports
import numpy as np

# local imports
from probeye.definition.noise_model import NormalNoiseModel
from probeye.inference.scipy_.correlation_models import (
    SpatialExponentialCorrelationModel,
)

# imports only needed for type hints
if TYPE_CHECKING:  # pragma: no cover
    from probeye.definition.sensor import Sensor


class NormalNoise(NormalNoiseModel):
    def __init__(
        self,
        target_sensor: "Sensor",
        prms_def: Union[str, List[Union[str, dict]], dict],
        name: Optional[str] = None,
        corr: Optional[str] = None,
        corr_model: Optional[str] = "exp",
        noise_type: str = "additive",
    ):
        """
        Parameters
        ----------
        target_sensor
            The sensor, the noise refers to. Note that this is not a sensor's name but
            a Sensor object. This object must contain the spatial information that is
            referred to in corr. For example, if corr is set to 'xy' then target_sensor
            must have a x and a y attribute that contain coordinate vectors of similar
            lengths.
        prms_def
            Contains the noise model's parameter names. The list may only contain
            strings or one-element dictionaries. It could look, for example, like
            [{'a': 'm'}, 'b']. The one-element dictionaries account for the possibility
            to define a local name for a latent parameter that is different from the
            global name. In the example above, the latent parameter with the global name
            'a' will be referred to as 'm' within the noise model. So, the one-element
            dicts have the meaning {<global name>: <local name>}. String-elements are
            interpreted as having similar local and global names. Note that the local-
            name option will not be required most of the times. The input from global
            to local name can also be provided as a dict. In the example above it would
            look like {'a': 'm', 'b': 'b'}.
        name
            Unique name of the noise model. This name is None, if the user does not
            specify it when adding the noise model to the problem. It is then named
            automatically before starting the inference engine.
        corr
            Spatial coordinates to be considered when setting up the covariance matrix.
            For example, when 'x' is chosen, the distance between two points i and j is
            computed as |xi - xj|. If 'xz' is chosen, this distance would be computed as
            sqrt((xi-xj)**2 + (zi-zj)**2). If None is chosen, no spatial correlation
            will be considered.
        corr_model
            Defines the correlation function the covariance matrix will be based on.
            Currently, there is only the one optional of an exponential model ('exp').
            More might be added in future versions. This argument has no effect if corr
            is set to None.
        noise_type
            Defines if the model error is computed by [prediction - measurement]
            ('additive') or via [prediction/measurement-1] ('multiplicative') or
            in some 'other' i.e., non-standard fashion.
        """

        # initialize the super-class (NormalNoiseModel) based on the given input
        super().__init__(
            prms_def=prms_def,
            sensors=target_sensor,
            name=name,
            corr=corr,
            corr_model=corr_model,
            noise_type=noise_type,
        )

        # the target_sensor is not an attribute of the super-class
        self.target_sensor = target_sensor

        # correlation related attributes
        self.corr = corr
        self.corr_model = corr_model
        if self.corr is not None:
            if self.corr_model == "exp":
                self.cov = SpatialExponentialCorrelationModel(
                    coords=target_sensor.coords, order=tuple(self.corr)
                )
                self.loglike_contribution = self.loglike_contribution_with_correlation  # type: ignore
            else:
                raise ValueError(
                    f"Encountered unknown flag '{self.corr_model}' for requested "
                    f"correlation model.\n Currently available options are: 'exp' for "
                    f"an exponential model."
                )
        else:
            self.loglike_contribution = self.loglike_contribution_without_correlation  # type: ignore

    def loglike_contribution_without_correlation(
        self, model_response_dict: dict, prms: dict, worst_value: float = -np.infty
    ) -> float:
        """
        This method overwrites the corresponding method of the parent class. Check out
        the docstring there for additional information.
        """
        # compute the model error; note that this mode has exactly one sensor
        model_error_vector = self.error(model_response_dict)[self.target_sensor.name]
        # the precision 'prec' is defined as the inverse of the variance, hence
        # prec = 1 / sigma**2 where sigma denotes the standard deviation
        std = prms["std"]
        if std < 0:
            return worst_value
        mean = 0.0 if self.zero_mean else prms["mean"]
        prec = 1.0 / std ** 2.0
        # evaluate the Gaussian log-PDF with zero mean and a variance of 1/prec for
        # each error term and sum them up
        ll = -len(model_error_vector) / 2 * np.log(2 * np.pi / prec)
        ll -= 0.5 * prec * np.sum(np.square(model_error_vector - mean))
        return ll

    def loglike_contribution_with_correlation(
        self, model_response_dict: dict, prms: dict, worst_value: float = -np.infty
    ) -> float:
        """
        Evaluates the log-likelihood for the noise model's experiments.

        Parameters
        ----------
        model_response_dict
            The first key is the name of the experiment. The values are dicts which
            contain the forward model's output sensor's names as keys have the
            corresponding model responses as values.
        prms
            Contains the names of the correlation model's parameters as keys and the
            corresponding numeric values as values.
        worst_value
            This value is returned when this method does not result in a numeric value.
            This might happen for example when the given parameters are not valid (for
            example in case of a negative standard deviation). The returned value in
            such cases should represent the worst possible value of the contribution.

        Returns
        -------
        ll
            The evaluated log-likelihood function.
        """
        # before computing the covariance matrix, check if the given parameter values
        # are valid; if not, the computation of the log-like contribution is stopped
        # right away while returning the worst value its evaluation can result in
        if not self.cov.check_prms(prms):
            return worst_value
        # since the covariance matrix is the same for all experiments in the noise
        # model, values derived from it (its inverse and determinant) are computed
        # before entering the loop over the noise model's experiments
        cov_matrix = self.cov(prms)
        inv_cov_matrix = np.linalg.inv(cov_matrix)
        # the following command computes the natural log and the sign of the matrix
        # cov_matrix; note that cov_matrix is positive definite, hence the sign will
        # always be positive and therefore isn't requested here
        _, log_det_cov_matrix = np.linalg.slogdet(cov_matrix)
        n = self.cov.n  # number of data points recorded by the sensor
        n_exp = len(self.experiment_names)
        # the term in the first parenthesis is added by each experiment in the loop;
        # this is why we can pre-compute it outside the loop
        ll = n_exp * (-(n * np.log(2 * np.pi) + log_det_cov_matrix) / 2)
        for exp_name in self.experiment_names:
            exp_dict = self.problem_experiments[exp_name]
            ym = model_response_dict[exp_name]
            ye = exp_dict["sensor_values"]
            error = self.error_function(ym, ye)[self.target_sensor.name]
            ll += -np.dot(error, inv_cov_matrix.dot(error)) / 2
        return ll


def translate_noise_model(noise_base: NormalNoiseModel) -> NormalNoise:
    """
    Translates a given instance of NoiseBase (which is essentially just a description
    of the noise model without computing-methods) to a specific noise model object
    which does contain computing-methods (e.g. compute the log-likelihood contribution).

    Parameters
    ----------
    noise_base
        An instance of NoiseBase which contains basic information on the noise model but
        no computing-methods.

    Returns
    -------
    noise_object
        An instance of a specific noise model class with computing capabilities.
        Examples for such classes are given above in this file.
    """

    # this is the noise classes currently defined (see code above)
    noise_classes = {"normal": NormalNoise}

    # this is where the translation happens
    noise_object = noise_classes[noise_base.dist](
        target_sensor=noise_base.sensors[0],
        prms_def=noise_base.prms_def,
        name=noise_base.name,
        corr=noise_base.corr,
        corr_model=noise_base.corr_model,
        noise_type=noise_base.noise_type,
    )

    # here, we take the assigned experiments from the base object
    noise_object.experiment_names = noise_base.experiment_names
    noise_object.problem_experiments = noise_base.problem_experiments

    return noise_object



# standard library
from typing import Union, List, Optional, Callable
import copy as cp

# third party imports
from tabulate import tabulate
from loguru import logger
import numpy as np
import torch as th

# local imports
from probeye.definition.parameter import Parameters
from probeye.definition.forward_model import ForwardModelBase
from probeye.definition.noise_model import NoiseModelBase
from probeye.subroutines import underlined_string, titled_table
from probeye.subroutines import simplified_list_string, simplified_dict_string
from probeye.subroutines import unvectorize_dict_values, make_list, len_or_one
from probeye.subroutines import print_probeye_header
from probeye.subroutines import logging_setup
from probeye.subroutines import add_index_to_tex_prm_name


class InferenceProblem:
    """
    This class provides a general framework for defining an inference problem.
    Capabilities for solving the set up problem are intentionally not included.
    """

    def __init__(
        self, name: str, use_default_logger: bool = True, log_file: Optional[str] = None
    ):
        """
        Parameters
        ----------
        name
            This is the name of the problem and has only descriptive value, for example
            when working with several inference problems.
        use_default_logger
            When True, the logger will be set up with some useful default values.
            Otherwise, no logger configurations are applied.
        log_file
            Path to the log-file, if the logging should be printed to file. If None is
            given, no logging-file will be created.
        """

        # the name of the problem
        self.name = name

        # this is the central parameter dictionary of the problem (the used Parameters-
        # class is derived from the dict-class); it contains all defined parameters
        # ('const' and 'latent' ones); the keys of this dictionary are the parameter
        # names; note that each parameter must have a unique name in the problem; the
        # values of this dictionary are Parameter-objects (see parameter.py) with the
        # following attributes:
        # .index  int or None (the index in the theta-vector (see self.loglike) for
        #         'latent'-parameter; None for 'const'-parameters)
        # .type   string (either 'model', 'prior' or 'noise' depending on where the
        #         parameter appears)
        # .prior  object or None (the prior-object of the 'latent'-parameter; None for
        #         'const'-parameters)
        # .value  float or None (defines the value for 'const'-parameters; None for
        #         'latent'-parameters)
        # .info   string (a short explanation of the parameter)
        # .tex:   string or None (the TeX version of the parameter's name, for example
        #         r'$\alpha$' for a parameter named 'alpha')
        self._parameters = Parameters()

        # this dictionary is intended for storing the measured data from experiment_
        # names (see self.add_experiment); this dict is managed internally and should
        # not be edited directly
        self._experiments = {}  # type: dict

        # here, the forward models are written to; note that the problem can have
        # multiple forward models; the keys are the forward model names, while the
        # values are the forward model objects, see also in the script forward_model.py;
        # this dictionary is managed internally and should not be edited directly
        self._forward_models = {}  # type: dict

        # a list for the problem's noise models (it's not a dict like the other
        # attributes from above as the noise models don't need to have names); this
        # list is managed internally and should not be edited directly
        self._noise_models = list()  # type: List[NoiseModelBase]

        # setup the logger with the given specifications
        if use_default_logger:
            logging_setup(log_file=log_file)

        # log probeye header and first message
        print_probeye_header()
        logger.debug("")  # for visual separation
        logger.debug(f"Initialized inference problem: '{self.name}'")

    @property
    def n_prms(self) -> int:
        """Provides n_prms attribute."""
        return self._parameters.n_prms

    @property
    def prms(self) -> List[str]:
        """Provides prms attribute."""
        return self._parameters.prms

    @property
    def n_latent_prms(self) -> int:
        """Provides n_latent_prms attribute."""
        return self._parameters.n_latent_prms

    @property
    def n_latent_prms_dim(self) -> int:
        """Provides n_latent_prms_dim attribute."""
        return self._parameters.n_latent_prms_dim

    @property
    def latent_prms(self) -> List[str]:
        """Provides latent_prms attribute."""
        return self._parameters.latent_prms

    @property
    def latent_prms_dims(self) -> List[int]:
        """Provides latent_prms_dims attribute."""
        return self._parameters.latent_prms_dims

    @property
    def n_constant_prms(self) -> int:
        """Provides n_constant_prms attribute."""
        return self._parameters.n_constant_prms

    @property
    def constant_prms(self) -> List[str]:
        """Provides constant_prms attribute."""
        return self._parameters.constant_prms

    @property
    def constant_prms_dict(self) -> dict:
        """Provides constant_prms_dict attribute."""
        return self._parameters.constant_prms_dict

    @property
    def n_model_prms(self) -> int:
        """Provides n_model_prms attribute."""
        return self._parameters.n_model_prms

    @property
    def model_prms(self) -> List[str]:
        """Provides model_prms attribute."""
        return self._parameters.model_prms

    @property
    def n_prior_prms(self) -> int:
        """Provides n_prior_prms attribute."""
        return self._parameters.n_prior_prms

    @property
    def prior_prms(self) -> List[str]:
        """Provides prior_prms attribute."""
        return self._parameters.prior_prms

    @property
    def n_noise_prms(self) -> int:
        """Provides n_noise_prms attribute."""
        return self._parameters.n_noise_prms

    @property
    def noise_prms(self) -> List[str]:
        """Provides noise_prms attribute."""
        return self._parameters.noise_prms

    @property
    def parameters(self) -> Parameters:
        """Access self._parameters from outside via self.parameters."""
        return self._parameters

    @property
    def priors(self) -> dict:
        """Provides the problem's prior-dictionary which is derived from the
        latent parameters in the self.parameters dictionary. The keys are the
        priors names, while the values are the prior-objects."""
        return {
            parameter_properties.prior.name: parameter_properties.prior
            for parameter_properties in self.parameters.values()
            if parameter_properties.is_latent
        }

    @property
    def noise_models(self) -> list:
        """Access self._noise_models from outside via self.noise_models."""
        return self._noise_models

    @property
    def forward_models(self) -> dict:
        """Access self._forward_models from outside via self.forward_models."""
        return self._forward_models

    @property
    def experiments(self) -> dict:
        """Access self._experiments from outside via self.experiments."""
        return self._experiments

    def info(
        self,
        include_experiments: bool = False,
        tablefmt: str = "presto",
        check_consistency: bool = True,
    ) -> str:
        """
        Either prints the problem definition to the console (print_it=True) or just
        returns the generated string without printing it (print_it=False).

        Parameters
        ----------
        include_experiments
            If True, information on the experiments defined within the model will be
            included in the printout. Depending on the number of defined experiments,
            this might result in a long additional printout, which is why this is set
            to False (no experiment printout) by default.
        tablefmt
            An argument for the tabulate function defining the style of the generated
            table. Check out tabulate's documentation for more info.
        check_consistency
            When True, a consistency check is performed before printing the explanations
            on theta. When False, this check is skipped.

        Returns
        -------
            The constructed string providing the problem information.
        """

        # contains the name of the inference problem
        title = f"Problem summary: {self.name}"
        title_string = underlined_string(title, n_empty_start=1)

        # list the forward models that have been defined within the problem
        rows = [
            (
                name,
                simplified_list_string([*model.prms_def.keys()]),
                simplified_list_string([*model.prms_def.values()]),
            )
            for name, model in self._forward_models.items()
        ]
        headers = ["Model name", "Global parameters", "Local parameters"]
        fwd_table = tabulate(rows, headers=headers, tablefmt=tablefmt)
        fwd_string = titled_table("Forward models", fwd_table)

        # include information on the defined priors
        rows = [
            (
                name,
                simplified_list_string([*prior.prms_def.keys()]),
                simplified_list_string([*prior.prms_def.values()]),
            )
            for name, prior in self.priors.items()
        ]
        headers = ["Prior name", "Global parameters", "Local parameters"]
        prior_table = tabulate(rows, headers=headers, tablefmt=tablefmt)
        prior_str = titled_table("Priors", prior_table)

        # provide various information on the problem's parameters
        prm_string = self._parameters.parameter_overview(tablefmt=tablefmt)
        prm_string += self._parameters.parameter_explanations(tablefmt=tablefmt)
        prm_string += self._parameters.const_parameter_values(tablefmt=tablefmt)

        # include the information on the theta interpretation
        theta_string = "\nTheta interpretation"
        theta_string += self.theta_explanation(check_consistency=check_consistency)

        # print information on added experiments if requested
        if include_experiments:
            rows_exp = []  # type: List[tuple[str, str]]
            for name, exp_dict in self._experiments.items():
                dict_atoms = unvectorize_dict_values(exp_dict["sensor_values"])
                for dict_atom in dict_atoms:
                    rows_exp.append((name, simplified_dict_string(dict_atom)))
            headers = ["Name", "Sensor values"]
            exp_table = tabulate(rows, headers=headers, tablefmt=tablefmt)
            exp_str = titled_table("Added experiments", exp_table)
        else:
            exp_str = ""

        # concatenate the string and return it
        full_string = title_string + fwd_string + prior_str + prm_string
        full_string += theta_string + exp_str

        # log and return the string
        for line in full_string.split("\n"):
            logger.info(line)
        return full_string

    def __str__(self) -> str:
        """
        Allows to print the problem definition via print(problem) if problem is
        an instance of InferenceProblem. See self.info for more details.
        """
        return self.info()

    def add_parameter(
        self,
        prm_name: str,
        prm_type: str,
        dim: Optional[int] = 1,
        const: Union[int, float, np.ndarray, None] = None,
        prior: Union[tuple, list, None] = None,
        info: str = "No explanation provided",
        tex: Optional[str] = None,
    ):
        """
        Adds a parameter ('const' or 'latent') to the inference problem. For more
        information, check out the Parameters.add_parameter method.
        """

        # add the parameter to the central parameter dictionary; checks and translations
        # are conducted in the Parameters.add_parameter method
        self._parameters.add_parameter(
            prm_name, prm_type, dim=dim, const=const, prior=prior, info=info, tex=tex
        )

    def remove_parameter(self, prm_name: str):
        """
        Removes a parameter ('const' or 'latent') from inference problem.

        Parameters
        ----------
        prm_name
            The name of the parameter to be removed.
        """

        # checks/additional actions are done by Parameters' __delitem__ method
        del self._parameters[prm_name]

    def change_parameter_role(
        self,
        prm_name: str,
        const: Union[int, float, None] = None,
        prior: Union[tuple, None] = None,
    ):
        """
        Performs the necessary tasks to change a parameter's role in the problem
        definition. A parameter's role can either be changed from 'const' to 'latent'
        or from 'latent' to 'const'.

        Parameters
        ----------
        prm_name
            The name of the parameter whose role should be changed.
        const
            If the new role is 'const', the corresponding value has to be specified by
            this argument.
        prior
            If the added parameter is a 'latent'-parameter, this argument has to be
            given as a 2-tuple. The first element (a string) defines the prior-type
            (will be referenced in inference routines). The 2nd element must be a
            dictionary stating the prior's parameters as keys and their numeric values
            as values or the name of a pre-defined parameter within the problem scope.
            An example for a normal prior: ('normal', {'loc': 0.0, 'scale': 1.0}). In
            order to define the prior's parameters, check out the prior definitions
            in priors.py.
        """
        # first, make sure that the given parameter exists
        self._parameters.confirm_that_parameter_exists(prm_name)

        # exactly one of the const and prior key word arguments must be given
        if const is not None and prior is not None:
            raise RuntimeError(
                f"You must specify either the 'const' or the 'prior' key argument. You "
                f"have specified both."
            )
        if const is None and prior is None:
            raise RuntimeError(
                f"You must specify either the 'const' or the 'prior' key argument. You "
                f"have specified none."
            )
        # raise an error if the role change would not change the role
        if self._parameters[prm_name].is_const and (prior is None):
            raise RuntimeError(
                f"The parameter '{prm_name}' is already defined as constant."
            )
        if self._parameters[prm_name].is_latent and (const is None):
            raise RuntimeError(
                f"The parameter '{prm_name}' is already defined as a "
                f"latent parameter."
            )
        # the parameter's role is changed by first removing it from the problem, and
        # then adding it again in its new role; the role-change does not impact the type
        # ('model', 'prior' or 'noise')
        prm = self._parameters[prm_name]
        if prm.is_const:
            # if a constant parameter should be made latent, its dimension will be taken
            # from its current value
            dim = len_or_one(prm.value)  # type: Union[int, None]
        else:
            # in this case, a latent parameter should be made constant; since the dim-
            # attribute will be inferred from the given value (const) it does not need
            # to be specified here
            dim = None
        self.remove_parameter(prm_name)
        self.add_parameter(
            prm_name,
            prm.type,
            dim=dim,
            const=const,
            prior=prior,
            info=prm.info,
            tex=prm.tex,
        )

    def change_parameter_info(
        self,
        prm_name: str,
        new_info: Optional[str] = None,
        new_tex: Optional[str] = None,
    ):
        """
        Changes the info-string and/or the tex-string of a given parameter.

        Parameters
        ----------
        prm_name
            The name of the parameter whose info-string should be changed.
        new_info
            The new string for the explanation of parameter prm_name.
        new_tex
            The new string for the parameter's tex-representation.
        """
        # first, make sure that the given parameter exists
        self._parameters.confirm_that_parameter_exists(prm_name)

        # if None is given for the new info/tex, the old value will be kept
        if new_info is None:
            new_info = self._parameters[prm_name].info
        if new_tex is None:
            new_tex = self._parameters[prm_name].tex

        # change the info/tex-string
        self._parameters[prm_name] = self._parameters[prm_name].changed_copy(
            info=new_info, tex=new_tex
        )

    def change_constant(self, prm_name: str, new_value: Union[int, float]):
        """
        Changes the value of a 'const'-parameter, i.e. a constant parameter of the
        inference problem.

        Parameters
        ----------
        prm_name
            The name of the 'const'-parameter whose value should be changed.
        new_value
            The new value that prm_name should assume.
        """
        # first, make sure that the given parameter exists
        self._parameters.confirm_that_parameter_exists(prm_name)

        # check if the given parameter is a constant
        if self._parameters[prm_name].is_latent:
            raise RuntimeError(f"The parameter '{prm_name}' is not a constant!")
        # change the parameter's value
        self._parameters[prm_name] = self._parameters[prm_name].changed_copy(
            value=new_value
        )

    def check_problem_consistency(self):
        """
        Conducts various checks to make sure the problem definition does not contain
        any inconsistencies.
        """

        # check if the central components have been added to the problem: parameters,
        # forward models, noise models and experiments
        assert (
            len(self._parameters) != 0
        ), "The problem does not contain any parameters!"
        assert (
            len(self._forward_models) != 0
        ), "The problem does not contain any forward models!"
        assert (
            len(self._noise_models) != 0
        ), "The problem does not contain any noise models!"
        assert (
            len(self._experiments) != 0
        ), "The problem does not contain any experiments!"

        # check if all parameters of the forward model(s) appear in self._parameters
        # and if they have the correct type
        for forward_model in self._forward_models.values():
            for model_prm in forward_model.prms_def.keys():
                assert model_prm in self._parameters, (
                    f"The forward model parameter '{model_prm}' is not defined "
                    f"within the problem!"
                )
                assert (
                    self._parameters[model_prm].type == "model"
                ), f"The forward model parameter '{model_prm}' is not of type 'model'!"

        # check if all parameters of the noise model appear in self._parameters
        # and if they have the correct type
        for noise_model in self._noise_models:
            for noise_prm in noise_model.prms_def.keys():
                assert noise_prm in self._parameters, (
                    f"The noise model parameter '{noise_prm}' is not defined within "
                    f"the problem!"
                )
                assert (
                    self._parameters[noise_prm].type == "noise"
                ), f"The noise model parameter '{noise_prm}' is not of type 'noise'!"

        # check if all prior objects in self.priors are consistent in terms of their
        # parameters; each one of them must appear in self._parameters
        for prior_obj in self.priors.values():
            for prior_prm in prior_obj.hyperparameters.keys():
                assert prior_prm in self._parameters, (
                    f"The prior parameter '{prior_prm}' is not defined within "
                    f"the problem!"
                )
                assert self._parameters[prior_prm].type == "prior", (
                    f"The prior parameter '{prior_prm}' " f"is not of type 'prior'!"
                )

        # check the indices of the latent parameters
        idx_list = []
        for prm_name, parameter in self._parameters.items():
            if parameter.is_latent:
                idx_range = [*range(parameter.index, parameter.index_end)]
                idx_list += idx_range
        assert sorted(idx_list) == list(range(len(idx_list))), (
            f"There seems to be an inconsistency in the latent parameter's indices. "
            f"The sorted index list is: "
            f"{sorted(idx_list)}"
        )

        # check the consistency of each parameter
        for parameter in self._parameters.values():
            parameter.check_consistency()

    def add_experiment(self, exp_name: str, sensor_values: dict, fwd_model_name: str):
        """
        Adds a single experiment to the inference problem. Here, an experiment is
        defined as one or more sensors (note that the experiment does not distinguish
        between input and output sensors) which provide some measured data.
        Additionally, a reference to one of the problem's forward models is associated
        with an experiment.

        Parameters
        ----------
        exp_name
            The name of the experiment, e.g. "Exp_20May.12". If an experiment with a
            similar name has already been added, it will be overwritten and a warning
            will be thrown.
        sensor_values
            The keys are the sensor's names, the values are the measured values.
        fwd_model_name
            Name of the forward model this experiment refers to.
        """

        # log at beginning so that errors can be associated
        logger.debug(f"Adding experiment '{exp_name}'")

        # check all keyword arguments are given
        if type(sensor_values) is not dict:
            raise TypeError(
                f"The sensor_values must be given as a dictionary. However, "
                f"found type '{type(sensor_values)}'."
            )
        if type(fwd_model_name) is not str:
            raise TypeError(
                f"The fwd_model_name must be given as a string. However, "
                f"found type '{type(fwd_model_name)}'."
            )

        # check if the given forward model exists
        if fwd_model_name not in self._forward_models.keys():
            raise RuntimeError(
                f"The forward model '{fwd_model_name}' does not exist! You need to "
                f"define it before adding experiments that refer to it."
            )

        # check that the stated forward model is consistent with the experiment
        experiment_sensors = [*sensor_values.keys()]
        input_sensors = self._forward_models[fwd_model_name].input_sensors
        for input_sensor in input_sensors:
            if input_sensor.name not in experiment_sensors:
                raise RuntimeError(
                    f"The forward model's ({fwd_model_name}) input sensor "
                    f"'{input_sensor.name}' is not provided by the given "
                    f"experiment '{exp_name}'!"
                )
        output_sensors = self._forward_models[fwd_model_name].output_sensors
        for output_sensor in output_sensors:
            if output_sensor.name not in experiment_sensors:
                raise RuntimeError(
                    f"The forward model's ({fwd_model_name}) output sensor "
                    f"'{output_sensor.name}' is not provided by the given "
                    f"experiment '{exp_name}'!"
                )

        # check if all sensor_values have the same lengths
        vector_lengths = set()
        for sensor_name, values in sensor_values.items():
            vector_lengths.add(len_or_one(values))
        if len(vector_lengths) > 1:
            raise RuntimeError(
                f"The sensor values must be all scalars or vectors of the same length. "
                f"However, found the lengths {vector_lengths}."
            )

        # check that vector-valued sensor_values are given as numpy-arrays; if not
        # (e.g. if lists or tuples are given) change them to numpy-ndarrays
        sensor_values_numpy = cp.copy(sensor_values)
        for sensor_name, values in sensor_values.items():
            if hasattr(values, "__len__"):
                if not isinstance(values, np.ndarray):
                    sensor_values_numpy[sensor_name] = np.array(values)

        # throw warning when the experiment name was defined before
        if exp_name in self._experiments.keys():
            logger.warning(
                f"Experiment '{exp_name}' is already defined "
                f"and will be overwritten!"
            )

        # add the experiment to the central dictionary
        self._experiments[exp_name] = {
            "sensor_values": sensor_values_numpy,
            "forward_model": fwd_model_name,
        }

    def get_parameters(
        self, theta: Union[np.ndarray, th.Tensor], prm_def: dict
    ) -> dict:
        """
        Extracts the numeric values for given parameters from the parameter vector theta
        and the constant parameters of the problem.

        Parameters
        ----------
        theta
            A numeric parameter vector passed to the loglike and logprior method. Which
            parameters these numbers refer to can be checked by calling
            self.theta_explanation() once the problem is set up.
        prm_def
            Defines which parameters to extract. The keys of this dictionary are the
            global parameter names, while the values are the local parameter names. In
            most cases global and local names will be identical, but sometimes it is
            convenient to define a local parameter name, e.g. in the forward model.

        Returns
        -------
        prms
            Contains <local parameter name> : <(global) parameter value> pairs.
        """
        prms = {}
        for global_name, local_name in prm_def.items():
            idx = self._parameters[global_name].index
            if idx is None:
                # in this case, the parameter is a constant and hence not read from
                # theta, but from the internal library
                prms[local_name] = self._parameters[global_name].value
            else:
                # in this case, the parameter is a latent parameter, and its value is
                # read from theta
                idx_end = self._parameters[global_name].index_end
                if idx == (idx_end - 1):
                    # scalars should not be returned as one-element-lists
                    prms[local_name] = theta[idx]
                else:
                    prms[local_name] = theta[idx:idx_end]
        return prms

    def get_experiment_names(
        self,
        forward_model_names: Union[str, List[str], None] = None,
        sensor_names: Union[str, List[str], None] = None,
        experiment_names: Union[str, List[str], None] = None,
    ) -> list:
        """
        Extracts the names of all experiments which refer to a given list of forward
        models and/or to a given list of sensor names from a given list of
        experiment names.

        Parameters
        ----------
        forward_model_names
            The names of the forward model the experiments should refer to.
        sensor_names
            The names of the sensors the experiments should should contain.
        experiment_names
            The names of the experiments to sub-select from. If None is given, then all
            experiments of the problem will be used.

        Returns
        -------
        relevant_experiment_names
            The names of the sub-selected experiments.
        """

        # at least one of forward_model_names and sensor_names must be given
        if (forward_model_names is None) and (sensor_names is None):
            raise RuntimeError(
                f"You did not specify any forward model(s) or sensor name(s)."
            )

        # if experiments is not further specified it is assumed that all given
        # experiments should be used
        if experiment_names is None:
            experiment_names = [*self._experiments.keys()]

        # this is for collecting the experiments
        relevant_experiment_names = []

        # get the experiments which refer to the given forward models
        if forward_model_names is not None:
            forward_model_names_list = make_list(forward_model_names)
            for exp_name in experiment_names:
                exp_dict = self._experiments[exp_name]
                fwd_model_name = exp_dict["forward_model"]
                if fwd_model_name in forward_model_names_list:
                    relevant_experiment_names.append(exp_name)
            experiment_names = relevant_experiment_names

        # get the experiments which contain the given sensors
        if sensor_names is not None:
            relevant_experiment_names = []
            sensor_names_list = make_list(sensor_names)
            for exp_name in experiment_names:
                exp_dict = self._experiments[exp_name]
                exp_sensors = [*exp_dict["sensor_values"].keys()]
                if all([s in exp_sensors for s in sensor_names_list]):
                    relevant_experiment_names.append(exp_name)

        return relevant_experiment_names

    def get_theta_names(self, tex: bool = False, components: bool = False) -> list:
        """
        Returns the parameter names of the parameter vector theta in the corresponding
        order (either in tex- or conventional format).

        Parameters
        ----------
        tex
            If True, the TeX-names of the parameters will be returned, otherwise the
            names as they are used in the code will be returned.
        components
            If True, parameters with dimension > 1 are returned component-wise; e.g.,
            if 'alpha' is a 2D parameter, it will be returned as 'alpha_1' and
            'alpha_2'. If False, only 'alpha' would be returned.

        Returns
        -------
        theta_names
            List of strings with the parameter names appearing in theta.
        """
        # assemble the parameter's names in the order as they appear in theta
        theta_names = []
        indices = []
        for prm_name, parameter in self._parameters.items():
            if parameter.is_latent:
                # add the parameter indices for checking later
                idx = parameter.index
                dim = parameter.dim
                if components and (dim > 1):
                    idx_end = parameter.index_end
                    indices += [*range(idx, idx_end)]
                else:
                    indices.append(parameter.index)
                # add the parameter names
                if tex and parameter.tex is not None:
                    if components and (dim > 1):
                        for i in range(dim):
                            theta_names.append(
                                add_index_to_tex_prm_name(parameter.tex, i + 1)
                            )
                    else:
                        theta_names.append(parameter.tex)
                else:
                    if components and (dim > 1):
                        for i in range(dim):
                            theta_names.append(f"{prm_name}_{i + 1}")
                    else:
                        theta_names.append(prm_name)
        # order the theta_names according to their index-values; note that this step is
        # not necessary for insertion ordered dicts (Python 3.6+), since in this case
        # theta_names will already be in the right order
        theta_names = [name for _, name in sorted(zip(indices, theta_names))]
        return theta_names

    def theta_explanation(self, check_consistency: bool = True) -> str:
        """
        Logs and returns a string on how the theta-vector, which is the numeric
        parameter vector that is given to the self.loglike and self.logprior methods,
        is interpreted with respect to the problem's parameters. The printout will tell
        you which parameter is connected to which index of theta.

        Parameters
        ----------
        check_consistency
            When True, a consistency check is performed before printing the
            explanations on theta. When False, this check is skipped.

        Returns
        -------
        s
            The constructed string when 'log' was set to False.
        """

        # an explanation is not printed if the problem is inconsistent
        if check_consistency:
            self.check_problem_consistency()

        # collect the list of theta names in the right order
        theta_names = self.get_theta_names()

        # construct the info-string
        s = "\n+---------------------------+\n"
        s += "|  Theta  |    Parameter    |\n"
        s += "|  index  |      name       |\n"
        s += "|---------------------------|\n"
        for prm_name in theta_names:
            idx = self.parameters[prm_name].index
            dim = self.parameters[prm_name].dim
            if dim == 1:
                s += f"|{idx:7d} --> {prm_name:<15s}|\n"
            else:
                idx_str = f"{idx}-{idx + dim - 1}"
                s += f"|{idx_str:>7s} --> {prm_name:<15s}|\n"
        s += "+---------------------------+\n"

        return s

    def add_forward_model(self, name: str, forward_model: ForwardModelBase):
        """
        Adds a forward model to the inference problem. Note that multiple forward models
        can be added to one problem.

        Parameters
        ----------
        name
            The name of the forward model to be added.
        forward_model
            Defines the forward model. Check out forward_model.py to see a template for
            the forward model definition. The user will then have to derive his own
            forward model from that base class.
        """

        # log at beginning so that errors can be associated
        logger.debug(f"Adding forward model '{name}'")

        # check if all given model parameters have already been added to the inference
        # problem; note that the forward model can only be added to the problem after
        # the corresponding parameters were defined
        for prm_name in forward_model.prms_def:
            self._parameters.confirm_that_parameter_exists(prm_name)

        # check if the given name for the forward model has already been used
        if name in [*self._forward_models.keys()]:
            raise RuntimeError(
                f"The given name '{name}' for the forward model has already been used "
                f"for another forward model. Please choose another name."
            )

        # check if the given forward model has an output sensor with a name that is
        # already used for an output sensor of another forward model
        for existing_name, existing_fwd_model in self._forward_models.items():
            for output_sensor in existing_fwd_model.output_sensor_names:
                if output_sensor in forward_model.output_sensor_names:
                    raise RuntimeError(
                        f"The given forward model '{name}' has an output sensor "
                        f"'{output_sensor}', \nwhich is also defined as an output "
                        f"sensor in the already defined forward model "
                        f"'{existing_name}'.\nPlease choose a different name for "
                        f"output sensor '{output_sensor}' in forward model '{name}'."
                    )

        # add the given forward model to the internal forward model dictionary under
        # the given forward model name
        self._forward_models[name] = forward_model

    def add_noise_model(self, noise_model: NoiseModelBase):
        """
        Adds a noise model to the inference problem.

        Parameters
        ----------
        noise_model
            The noise model object, e.g. from NormalNoise. Check out noise.py to see
            some noise model classes.
        """

        # check if the noise model has been assigned a name; if not, assign one
        if noise_model.name is None:
            noise_model.name = f"noise_model_{len(self.noise_models)}"
            logger.debug(
                f"Adding noise model '{noise_model.name}' (name assigned automatically)"
            )
        else:
            logger.debug(f"Adding noise model '{noise_model.name}'")

        # check if all given noise model parameters have already been added to
        # the inference problem
        for prm_name in noise_model.prms_def:
            if prm_name not in self._parameters.keys():
                raise RuntimeError(
                    f"The noise model parameter '{prm_name}' has not been defined yet."
                    f"\nYou have to add all noise model parameters to the problem "
                    f"before adding the noise model.\nYou can use the 'add_parameter' "
                    f"method for this purpose."
                )

        # add the problem's experiments to the noise model (this is just a pointer!)
        # for noise_model-internal checks
        noise_model.problem_experiments = self._experiments
        # finally, add the noise_model to the internal dict
        self._noise_models.append(noise_model)

    def assign_experiments_to_noise_models(self):
        """
        Assigns each noise model the corresponding experiment names, based on the sensor
        names, that are defined for each noise model. This function is intended to be
        called after the problem was fully defined. Alternatively, you can assign the
        experiments 'by hand', by using the NoiseModelBase's 'add_experiments' method.
        """
        logger.debug("Assigning experiments to noise models")
        n_experiments_defined = len(self._experiments)
        n_experiments_noise = 0
        for noise_model in self._noise_models:
            # get the experiments that contain all of the noise model's sensors
            experiment_names = self.get_experiment_names(
                sensor_names=noise_model.sensor_names
            )
            n_experiments_noise += len(experiment_names)
            # add the relevant experiment names to the noise model
            noise_model.add_experiments(experiment_names)

        # check if there is the same number of experiments over all noise models as
        # defined for the inference problem
        if n_experiments_noise != n_experiments_defined:
            # this is not necessarily an error; it also happens in a valid setup when
            # more than one noise model are defined for one forward model; in a future
            # version, there could be an info message here
            pass

        # check that each globally defined experiment appears in one of the noise models
        exp_names_in_noise_models = set()
        for noise_model in self._noise_models:
            for exp_name_noise in noise_model.experiment_names:
                exp_names_in_noise_models.add(exp_name_noise)

        for exp_name in self._experiments.keys():
            if exp_name not in exp_names_in_noise_models:
                # undo the adding of the experiment names
                for noise_model in self._noise_models:
                    noise_model.experiment_names = []
                # one may argue, that this could also be only a warning here
                raise RuntimeError(
                    f"The globally defined experiment '{exp_name}' does not appear in "
                    f"any of the noise models!"
                )

    def transform_experimental_data(
        self, f: Callable, args: tuple = (), **kwargs
    ) -> "InferenceProblem":
        """
        Creates a full copy of the problem the experimental data of which is transformed
        in some way. This might be a necessary pre-processing step for an inference
        engine in order to be able to solve the problem. Note that the original problem
        remains unchanged.

        Parameters
        ----------
        f
            The function that is applied on each of the experiment's sensor values.
        args
            Additional positional arguments to be passed to f.
        kwargs
            Keyword arguments to be passed to f.

        Returns
        -------
        self_copy
            A full copy of self where the experimental data has been transformed in the
            specified fashion.
        """

        logger.debug(f"Transforming experimental data using f = '{f.__name__}'")

        # the original problem shall not be touched, so we create a copy here to which
        # the transformation will be applied; however, sometimes this is not possible,
        # hence the try-except frame
        try:
            self_copy = cp.deepcopy(self)
        except:
            logger.warning(
                "The inference problem could not be deep-copied! "
                "The original problem will be modified!"
            )
            self_copy = self

        # transform the sensor values from the experiments by applying the specified
        # function with the given arguments to them
        for exp_name in self_copy._experiments.keys():
            sensor_values = self_copy._experiments[exp_name]["sensor_values"]
            for sensor_name in sensor_values.keys():
                sensor_values[sensor_name] = f(
                    sensor_values[sensor_name], *args, **kwargs
                )

        return self_copy



# standard library imports
import unittest

# local imports
from probeye.definition.parameter import Parameters
from probeye.definition.parameter import ParameterProperties
from probeye.definition.prior import PriorBase


class TestProblem(unittest.TestCase):
    def test_parameters(self):
        # this is the standard way to use Parameters
        parameters = Parameters()
        parameters["a"] = ParameterProperties(
            {
                "index": None,
                "type": "model",
                "prior": None,
                "value": 1.0,
                "info": "...",
                "tex": r"$a$",
            }
        )
        parameters["b"] = ParameterProperties(
            {
                "index": None,
                "type": "model",
                "prior": None,
                "value": 2.0,
                "info": "...",
                "tex": r"$b$",
            }
        )
        c_prior = PriorBase("c", ["s"], "c_dummy", "normal")
        parameters["c"] = ParameterProperties(
            {
                "index": 2,
                "dim": 1,
                "type": "model",
                "prior": c_prior,
                "value": None,
                "info": "...",
                "tex": r"$c$",
            }
        )
        # check invalid key
        with self.assertRaises(ValueError):
            parameters[True] = ParameterProperties(
                {
                    "index": None,
                    "type": "model",
                    "prior": None,
                    "value": 3.0,
                    "info": "...",
                    "tex": r"$d$",
                }
            )
        # check if the role is correctly identified
        self.assertEqual(parameters["a"].role, "const")
        self.assertEqual(parameters["c"].role, "latent")
        # check for the parameter-role bookkeeping
        self.assertEqual(parameters.n_latent_prms, 1)
        self.assertEqual(parameters.n_constant_prms, 2)
        self.assertEqual(parameters.n_prms, 3)
        # you cannot add the parameter's properties as a dictionary
        with self.assertRaises(ValueError):
            d_prior = PriorBase("d", ["s"], "d_dummy", "normal")
            parameters["d"] = {
                "index": 3,
                "dim": 1,
                "type": "model",
                "role": "latent",
                "prior": d_prior,
                "value": None,
                "info": "...",
                "tex": r"$d$",
            }

        # you should not be able to set parameter attributes directly
        with self.assertRaises(AttributeError):
            parameters["a"].index = -1
        with self.assertRaises(AttributeError):
            parameters["a"].dim = -1
        with self.assertRaises(AttributeError):
            parameters["a"].type = -1
        with self.assertRaises(AttributeError):
            parameters["a"].role = -1
        with self.assertRaises(AttributeError):
            parameters["a"].prior = -1
        with self.assertRaises(AttributeError):
            parameters["a"].value = -1

        # check if the consistency-checks work
        with self.assertRaises(TypeError):
            # index is has wrong type
            parameters["d"] = ParameterProperties(
                {
                    "index": True,
                    "dim": 1,
                    "type": "model",
                    "prior": None,
                    "value": None,
                    "info": "...",
                    "tex": r"$d$",
                }
            )
        with self.assertRaises(ValueError):
            # index has invalid value
            parameters["d"] = ParameterProperties(
                {
                    "index": -1,
                    "dim": 1,
                    "type": "model",
                    "prior": None,
                    "value": 1.0,
                    "info": "...",
                    "tex": r"$d$",
                }
            )
        with self.assertRaises(TypeError):
            # dim is has wrong type
            d_prior = PriorBase("d", ["s"], "d_dummy", "normal")
            parameters["d"] = ParameterProperties(
                {
                    "index": 3,
                    "dim": True,
                    "type": "model",
                    "prior": d_prior,
                    "value": None,
                    "info": "...",
                    "tex": r"$d$",
                }
            )
        with self.assertRaises(ValueError):
            # dim has invalid value
            d_prior = PriorBase("d", ["s"], "d_dummy", "normal")
            parameters["d"] = ParameterProperties(
                {
                    "index": 3,
                    "dim": 0,
                    "type": "model",
                    "prior": d_prior,
                    "value": None,
                    "info": "...",
                    "tex": r"$d$",
                }
            )
        with self.assertRaises(RuntimeError):
            # type has invalid value
            parameters["d"] = ParameterProperties(
                {
                    "index": None,
                    "type": "invalid value",
                    "prior": None,
                    "value": 1.0,
                    "info": "...",
                    "tex": r"$d$",
                }
            )
        with self.assertRaises(TypeError):
            # prior has invalid type
            parameters["d"] = ParameterProperties(
                {
                    "index": None,
                    "type": "model",
                    "prior": True,
                    "value": 1.0,
                    "info": "...",
                    "tex": r"$d$",
                }
            )
        with self.assertRaises(TypeError):
            # value has invalid type
            parameters["d"] = ParameterProperties(
                {
                    "index": None,
                    "type": "model",
                    "prior": None,
                    "value": True,
                    "info": "...",
                    "tex": r"$d$",
                }
            )
        with self.assertRaises(RuntimeError):
            # index and dim are not consistent
            d_prior = PriorBase("d", ["s"], "d_dummy", "normal")
            parameters["d"] = ParameterProperties(
                {
                    "index": 3,
                    "dim": None,
                    "type": "model",
                    "prior": d_prior,
                    "value": None,
                    "info": "...",
                    "tex": r"$d$",
                }
            )
        with self.assertRaises(RuntimeError):
            # index and prior are inconsistently combined
            parameters["d"] = ParameterProperties(
                {
                    "index": 3,
                    "dim": 1,
                    "type": "model",
                    "prior": None,
                    "value": None,
                    "info": "...",
                    "tex": r"$d$",
                }
            )
        with self.assertRaises(RuntimeError):
            # index and value are inconsistently combined
            d_prior = PriorBase("d", ["s"], "d_dummy", "normal")
            parameters["d"] = ParameterProperties(
                {
                    "index": 3,
                    "dim": 1,
                    "type": "model",
                    "prior": d_prior,
                    "value": 1.0,
                    "info": "...",
                    "tex": r"$d$",
                }
            )
        with self.assertRaises(RuntimeError):
            # index and prior are inconsistently combined
            d_prior = PriorBase("d", ["s"], "d_dummy", "normal")
            parameters["d"] = ParameterProperties(
                {
                    "index": None,
                    "type": "model",
                    "prior": d_prior,
                    "value": 1.0,
                    "info": "...",
                    "tex": r"$d$",
                }
            )
        with self.assertRaises(RuntimeError):
            # index and value are inconsistently combined
            parameters["d"] = ParameterProperties(
                {
                    "index": None,
                    "type": "model",
                    "prior": None,
                    "value": None,
                    "info": "...",
                    "tex": r"$d$",
                }
            )


if __name__ == "__main__":
    unittest.main()



"""This module contains functions that remove version number constraints from
selected packages in the setup.cfg file. It is meant to be used during CI: to test if
the package works with the latest versions of its dependencies."""

import configparser
import re
from typing import List


def version_constraint_free_packages(setup_cfg_packages: str) -> List[str]:
    """
    Remove version constraints from the packages listed in `options_field` (provided
    as input argument or read from `setup.cfg`).

    Parameters
    ----------
    setup_cfg_packages: str
        Packages with optional version constraints as parsed from a syntactically
        correct `setup.cfg` file.

    Returns
    -------
    List[str]
        `install_requires` packages without version constraints.
    """
    # to be able to match the last package in all cases
    setup_cfg_packages = "\n" + setup_cfg_packages + "\n"

    # get plain (version constraint free) package names, they are listed using a
    # "list-semi": dangling list or string of semicolon-separated values
    # https://setuptools.pypa.io/en/latest/userguide/declarative_config.html#options
    packages_without_version = re.findall(
        r"(?<=(?:\n|;)).*?(?=<|=|>|!|;|\n|;])", setup_cfg_packages
    )
    # strip possible remaining spaces and drop empty string elements
    packages_without_version = [s.strip() for s in packages_without_version]
    packages_without_version = list(filter(None, packages_without_version))

    return packages_without_version


def version_constraint_free_dependencies(options_field: str) -> None:
    """
    Remove version constraints from the packages listed in `options_field` (
    provided as input argument or read from `setup.cfg`) and overwrite the
    `setup.cfg` file with the version constraint-free dependencies. Useful in a CI
    pipeline to test if the package works with the latest versions of its dependencies.

    Parameters
    ----------
    options_field: str
        Field name in `setup.cfg` options that list package dependencies,
        e.g. `"install_requires"`. For further information see:
        https://setuptools.pypa.io/en/latest/userguide/declarative_config.html#options

    """
    config = configparser.ConfigParser()
    config.read("setup.cfg")
    packages = config["options"][options_field]

    packages_without_version = version_constraint_free_packages(packages)
    setup_cfg_packages_without_version = "\n" + "\n".join(packages_without_version)

    config["options"][options_field] = setup_cfg_packages_without_version

    with open("setup.cfg", "w") as configfile:
        config.write(configfile)


# to ease running the version number constraint removal from the command line
if __name__ == "__main__":
    # this is meant to be run only during CI: to test if the package works with the
    # latest versions of its dependencies
    version_constraint_free_dependencies(options_field="install_requires")



# standard library imports
from typing import Union, List, Dict

# third party imports
import pyro
import pyro.distributions as dist
import torch as th

# local imports
from probeye.definition.prior import PriorBase
from probeye.subroutines import len_or_one


class PriorNormal(PriorBase):
    """Prior class for a normal distribution."""

    def __init__(
        self,
        ref_prm: str,
        prms_def: Union[str, dict, List[Union[str, dict]]],
        name: str,
    ):
        """
        Parameters
        ----------
        ref_prm
            The name of the latent parameter the prior refers to.
        prms_def
            Defines the prior's parameter names. See the docstring of PriorBase for
            more detailed information.
        name
            Defining the priors name.
        """
        super().__init__(ref_prm, prms_def, name, "normal distribution")

    def sample(self, prms: dict) -> pyro.sample:
        """
        Samples from a (multivariate-)normal distribution.

        Parameters
        ----------
        prms
            Contains the prior's parameters as keys and their values as values.

        Returns
        -------
            The generated sample.
        """
        loc = prms[f"loc_{self.ref_prm}"]
        scale = prms[f"scale_{self.ref_prm}"]
        if len_or_one(loc) == 1:
            distribution = dist.Normal(loc, scale)
        else:
            distribution = dist.MultivariateNormal(
                th.from_numpy(loc), covariance_matrix=th.from_numpy(scale)
            )
        return pyro.sample(self.ref_prm, distribution)


class PriorUniform(PriorBase):
    """Prior class for a uniform distribution."""

    def __init__(
        self,
        ref_prm: str,
        prms_def: Union[str, dict, List[Union[str, dict]]],
        name: str,
    ):
        """
        Parameters
        ----------
        ref_prm
            The name of the latent parameter the prior refers to.
        prms_def
            Defines the prior's parameter names. See the docstring of PriorBase for
            more detailed information.
        name
            Defining the priors name.
        """
        super().__init__(ref_prm, prms_def, name, "uniform distribution")

    def sample(self, prms: dict) -> pyro.sample:
        """
        Samples from a (multivariate-)normal distribution.

        Parameters
        ----------
        prms
            Contains the prior's parameters as keys and their values as values.

        Returns
        -------
            The generated sample.
        """
        low = prms[f"low_{self.ref_prm}"]
        high = prms[f"high_{self.ref_prm}"]
        return pyro.sample(self.ref_prm, dist.Uniform(low, high))


def translate_prior_template(prior_template: PriorBase) -> PriorBase:
    """
    Translate a given instance of PriorBase (which is essentially just a description of
    the prior without compute-methods) to a specific prior object which does contain
    compute-methods (e.g. compute the logpdf).

    Parameters
    ----------
    prior_template
        An instance of PriorBase which contains basic information on the prior but no
        computing-methods.

    Returns
    -------
    prior_object
        An instance of a specific prior class with computing capabilities. Examples for
        such classes are given above in this file.
    """

    # this dictionary contains the classes from above in this file
    prior_classes = {"normal": PriorNormal, "uniform": PriorUniform}

    # prepare the corresponding prior object; the following translation is necessary,
    # because prms_def must be given in form of a list, but was already translated to a
    # dictionary when instantiating the PriorBase objects; hence prior_template.prms_def
    # is a dictionary
    prms_def = [
        {key: value}
        for key, value in prior_template.prms_def.items()
        if key != prior_template.ref_prm
    ]
    prior_object = prior_classes[prior_template.prior_type](
        prior_template.ref_prm, prms_def, prior_template.name  # type: ignore
    )

    return prior_object



# standard library imports
import unittest

# third party imports
import numpy as np

# local imports
from probeye.inference.scipy_.correlation_models import (
    SpatialExponentialCorrelationModel,
)


class TestProblem(unittest.TestCase):
    def test_check_prms(self):
        x_test = np.linspace(0.0, 1.0, 10)
        scm = SpatialExponentialCorrelationModel(x=x_test)
        self.assertTrue(not scm.check_prms({"l_corr": -1.0, "std": 1.0}))
        self.assertTrue(not scm.check_prms({"l_corr": 1.0, "std": -1.0}))


if __name__ == "__main__":
    unittest.main()



# standard library
import logging
import unittest

# third party imports
import numpy as np

# local imports
from probeye.definition.forward_model import ForwardModelBase
from probeye.definition.sensor import Sensor
from probeye.definition.inference_problem import InferenceProblem
from probeye.definition.noise_model import NormalNoiseModel
from probeye.inference.dynesty_.solver import DynestySolver


class TestProblem(unittest.TestCase):
    def test_dynesty_solver(self):

        np.random.seed(6174)

        # define the forward model
        class LinRe(ForwardModelBase):
            def __call__(self, inp):
                x = inp["x"]
                a = inp["a"]
                b = inp["b"]
                return {"y": a * x + b}

        # set up the problem
        problem = InferenceProblem("Linear regression")
        problem.add_parameter("a", "model", prior=("normal", {"loc": 0, "scale": 1}))
        problem.add_parameter("b", "model", prior=("normal", {"loc": 0, "scale": 1}))
        problem.add_parameter(
            "sigma", "noise", prior=("uniform", {"low": 0.1, "high": 1})
        )
        problem.add_forward_model(
            "LinRe", LinRe(["a", "b"], [Sensor("x")], [Sensor("y")])
        )
        problem.add_noise_model(NormalNoiseModel({"sigma": "std"}, sensors=Sensor("y")))

        # generate and add some simple test data
        n_tests = 5000
        true = {"a": 0.3, "b": -0.2, "sigma": 0.1}
        x_test = np.linspace(0.0, 1.0, n_tests)
        y_true = true["a"] * x_test + true["b"]
        y_test = np.random.normal(loc=y_true, scale=true["sigma"])
        problem.add_experiment(
            f"Tests", fwd_model_name="LinRe", sensor_values={"x": x_test, "y": y_test}
        )

        # run the dynesty solver with deactivated output
        logging.root.disabled = True
        dynesty_solver = DynestySolver(problem, show_progress=True, seed=6174)
        dynesty_solver.run_dynesty("dynamic", nlive_init=10, nlive_batch=10, maxbatch=2)

        sample_means = dynesty_solver.summary["mean"]
        for parameter, true_value in true.items():
            self.assertAlmostEqual(sample_means[parameter], true_value, delta=0.01)


if __name__ == "__main__":
    unittest.main()



"""
Linear regression example solved with max. likelihood (without needing priors).
----------------------------------------------------------------------------------------
The model equation is y = m * x + b with a, b being the model parameters and the noise
model is a normal zero-mean distribution with the std. deviation to infer. The problem
is solved via maximum likelihood estimation based on scipy.
"""

# standard library imports
import unittest

# third party imports
import numpy as np
import matplotlib.pyplot as plt

# local imports (problem definition)
from probeye.definition.inference_problem import InferenceProblem
from probeye.definition.forward_model import ForwardModelBase
from probeye.definition.sensor import Sensor
from probeye.definition.noise_model import NormalNoiseModel

# local imports (inference related)
from probeye.inference.scipy_.solver import ScipySolver


class TestProblem(unittest.TestCase):
    def test_maximum_likelihood(self, plot: bool = False):
        """
        Integration test for the problem described at the top of this file.

        Parameters
        ----------
        plot
            If True, the problem's data is plotted. This is deactivated by default, so
            that the test does not stop until the generated plots are closed.
        """

        # ============================================================================ #
        #                              Set numeric values                              #
        # ============================================================================ #

        # 'true' value of m
        m_true = 2.5

        # 'true' value of b
        b_true = 1.7

        # 'true' value of noise sd
        sigma_noise = 0.5

        # the number of generated experiment_names and seed for random numbers
        n_tests = 50
        seed = 1

        # ============================================================================ #
        #                           Define the Forward Model                           #
        # ============================================================================ #

        class LinearModel(ForwardModelBase):
            def response(self, inp: dict) -> dict:
                # this method *must* be provided by the user
                x = inp["x"]
                m = inp["m"]
                b = inp["b"]
                response = {}
                for os in self.output_sensors:
                    response[os.name] = m * x + b
                return response

        # ============================================================================ #
        #                         Define the Inference Problem                         #
        # ============================================================================ #

        # initialize the inference problem with a useful name; note that the name will
        # only be stored as an attribute of the InferenceProblem and is not important
        # for the problem itself; can be useful when dealing with multiple problems
        problem = InferenceProblem("Max likelihood for linear regression")

        # add all parameters to the problem; the first argument states the parameter's
        # global name (here: 'm', 'b' and 'sigma'); the second argument defines the
        # parameter type (three options: 'model' for parameter's of the forward model,
        # 'prior' for prior parameters and 'noise' for parameters of the noise model);
        # the tex argument is states a tex-string for the parameter which is only used
        # for plotting
        problem.add_parameter("m", "model", tex="$m$")
        problem.add_parameter("b", "model", tex="$b$")
        problem.add_parameter("sigma", "noise", tex=r"$\sigma$")

        # add the forward model to the problem
        isensor = Sensor("x")
        osensor = Sensor("y")
        linear_model = LinearModel(["m", "b"], [isensor], [osensor])
        problem.add_forward_model("LinearModel", linear_model)

        # add the noise model to the problem
        problem.add_noise_model(
            NormalNoiseModel(prms_def={"sigma": "std"}, sensors=osensor)
        )

        # ============================================================================ #
        #                    Add test data to the Inference Problem                    #
        # ============================================================================ #

        # data-generation; normal noise with constant variance around each point
        np.random.seed(seed)
        x_test = np.linspace(0.0, 1.0, n_tests)
        y_true = linear_model.response(
            {isensor.name: x_test, "m": m_true, "b": b_true}
        )[osensor.name]
        y_test = np.random.normal(loc=y_true, scale=sigma_noise)

        # add the experimental data
        problem.add_experiment(
            f"TestSeries_1",
            fwd_model_name="LinearModel",
            sensor_values={isensor.name: x_test, osensor.name: y_test},
        )

        # give problem overview
        problem.info()

        # plot the true and noisy data
        if plot:
            plt.scatter(x_test, y_test, label="measured data", s=10, c="red", zorder=10)
            plt.plot(x_test, y_true, label="true", c="black")
            plt.xlabel(isensor.name)
            plt.ylabel(osensor.name)
            plt.legend()
            plt.tight_layout()
            plt.draw()  # does not stop execution

        # ============================================================================ #
        #                    Solve problem with inference engine(s)                    #
        # ============================================================================ #

        # this routine is imported from another script because it it used by all
        # integration tests in the same way; ref_values are used for plotting
        true_values = {"m": m_true, "b": b_true, "sigma": sigma_noise}
        scipy_solver = ScipySolver(problem)
        scipy_solver.run_max_likelihood(true_values=true_values)


if __name__ == "__main__":
    unittest.main()



# standard library imports
from typing import Union, List, Tuple, Optional, Callable, Iterable, TYPE_CHECKING, Any
import warnings

# third party imports
import numpy as np
import pyro.infer.mcmc.util
import torch as th
from pyro.infer import NUTS, MCMC
import arviz as az
from loguru import logger
import time
import contextlib

# local imports
from probeye.inference.torch_.priors import translate_prior_template
from probeye.inference.torch_.noise_models import translate_noise_model
from probeye.subroutines import len_or_one, make_list
from probeye.subroutines import pretty_time_delta, stream_to_logger
from probeye.subroutines import print_dict_in_rows
from probeye.subroutines import check_for_uninformative_priors

# imports only needed for type hints
if TYPE_CHECKING:  # pragma: no cover
    from probeye.definition.inference_problem import InferenceProblem
    from probeye.definition.forward_model import ForwardModelBase


class PyroSolver:
    """Solver routines based on pyro/torch for an InferenceProblem."""

    def __init__(
        self, problem: "InferenceProblem", seed: int = 1, show_progress: bool = True
    ):
        """
        Parameters
        ----------
        problem
            Describes the inference problem including e.g. parameters and data.
        seed
            Random state used for random number generation.
        show_progress
            When True, the progress of a solver routine will be shown (for example as a
            progress-bar) if such a feature is available. Otherwise, the progress will
            not shown.
        """

        # check that the problem does not contain a uninformative prior
        check_for_uninformative_priors(problem)

        # attributes from arguments
        self.show_progress = show_progress
        self.seed = seed

        # the following attributes will be set after the solver was run
        self.raw_results = None
        self.summary = {}  # type: dict

        # the problem is copied, and in the copy, the experimental data is reformatted
        # from numpy-arrays to torch-tensors; note that the first command makes sure
        # that also scalars are converted to numpy-arrays
        self.problem = problem.transform_experimental_data(f=np.atleast_1d)
        self.problem = self.problem.transform_experimental_data(f=th.from_numpy)

        # each noise model must be connected to the relevant experiment_names
        self.problem.assign_experiments_to_noise_models()

        # the dictionary dependency_dict will contain all latent parameter names as
        # keys; the value of each key will be a list with latent hyper-parameters of the
        # latent parameter's prior; note that the dependency_dict is made an attribute,
        # so that one can better test those routines
        logger.debug("Checking parameter's dependencies")
        self.dependency_dict = dict()  # type: dict
        for prm_name in self.problem.parameters.latent_prms:
            self.dependency_dict[prm_name] = []
            hyperparameters = self.problem.parameters[prm_name].prior.hyperparameters
            for prior_prm_name in hyperparameters:
                if prior_prm_name in self.problem.parameters.latent_prms:
                    self.dependency_dict[prm_name].append(prior_prm_name)

        # this makes sure that the items in dependency are in an order that they can be
        # sampled from beginning (index 0) sequentially until the last item without
        # encountering any dependency problems (i.e. that another parameter has to be
        # sampled before the considered prm can be sampled)
        consistent = False
        while not consistent:
            consistent = True
            idx_latent_dependencies = [
                i for i, v in enumerate(self.dependency_dict.values()) if len(v) > 0
            ]
            for idx in idx_latent_dependencies:
                key_idx = [*self.dependency_dict.keys()][idx]
                for dependency in self.dependency_dict[key_idx]:
                    if key_idx in self.dependency_dict[dependency]:
                        raise RuntimeError(
                            f"Found circular dependency between {key_idx} and "
                            f"{dependency}!"
                        )
                    idx_dependency = [*self.dependency_dict.keys()].index(dependency)
                    if idx_dependency > idx:
                        consistent = False
                        tuples = [*self.dependency_dict.items()]
                        tuples[idx], tuples[idx_dependency] = (
                            tuples[idx_dependency],
                            tuples[idx],
                        )
                        self.dependency_dict = dict(tuples)

        # translate the prior definitions to objects with computing capabilities
        logger.debug("Translating problem's priors")
        self.priors = {}
        for prm_name in self.dependency_dict:
            prior_template = self.problem.parameters[prm_name].prior
            self.priors[prior_template.ref_prm] = translate_prior_template(
                prior_template
            )

        # translate the general noise model objects into solver specific ones
        logger.debug("Translating problem's noise models")
        self.noise_models = []
        for noise_model_base in self.problem.noise_models:
            self.noise_models.append(translate_noise_model(noise_model_base))

        # translate the problem's forward models into torch compatible ones
        logger.debug("Wrapping problem's forward models")
        self.wrapped_forward_models = {}
        for fwd_model_name in self.problem.forward_models:
            self.wrapped_forward_models[fwd_model_name] = self._translate_forward_model(
                self.problem.forward_models[fwd_model_name]
            )

    @staticmethod
    def _only_values(func: Callable) -> Callable:
        """
        This function wrapper is required for the Autograd.apply function which is
        returned by translate_forward_model. While the forward model's response method
        takes a dict as input, the Autograd.apply function which will essentially take
        the forward model's response method's place requires the numeric values of inp
        given as positional arguments. This pre-processing step is done by this function
        wrapper.
        """

        def wrapper(inp: dict):
            return func(*inp.values())

        return wrapper

    def _translate_forward_model(self, forward_model: "ForwardModelBase") -> Callable:
        """
        Translates a given forward model (based on non-tensor in/outputs) to a torch-
        compatible forward model based on tensors.

        Parameters
        ----------
        forward_model
            The forward model object. Note that this object needs to have a jacobian
            method to make it translatable.
        """

        class Autograd(th.autograd.Function):
            @staticmethod
            def forward(ctx: Any, *values: th.Tensor) -> th.Tensor:  # type: ignore
                """
                Modified from the th.autograd.Function docstring: This function must
                accept a context ctx as the first argument, followed by any number of
                arguments (tensors or other types). The context can be used to store
                arbitrary data that can be then retrieved during the backward pass.

                Parameters
                ----------
                ctx
                    This context-object is for derivative-bookkeeping.
                values
                    The values of the inp-dict of the inp-argument of the forward
                    model's 'response' method in the same order.
                """

                # the idea is that the given forward model only takes numeric
                # (non-tensor) inputs; so there needs to be a conversion which takes
                # place here; also, the given values must be rearranged in the dict-
                # format required by forward_model's response method
                inp = {}
                keys = forward_model.input_channel_names
                for key, value in zip(keys, values):
                    if th.is_tensor(value):
                        inp[key] = value.detach().numpy()
                    else:
                        inp[key] = value
                    forward_model.input_structure[key] = len_or_one(inp[key])

                # evaluate the forward model and its jacobian for the given input
                # parameters; this is where the forward model is evaluated in its
                # original setup without any tensors
                response_dict = forward_model.response(inp)
                jac_dict = forward_model.jacobian(inp)
                jac_numpy = forward_model.jacobian_dict_to_array(
                    inp, jac_dict, self.problem.n_latent_prms_dim
                )

                # now we have the forward model's response in dict format; however, in
                # this format it cannot be processed here, so we will put all its
                # numeric results in one long vector
                n1 = len(forward_model.output_sensors)
                n1_times_n3, _ = jac_numpy.shape
                n3 = n1_times_n3 // n1
                response_numpy = np.zeros(n1_times_n3)
                response_structure = dict()
                for i, (key, value) in enumerate(response_dict.items()):
                    n_elements = len_or_one(value)
                    idx_start = i * n3
                    idx_end = idx_start + n_elements
                    response_numpy[idx_start:idx_end] = value
                    response_structure[key] = n_elements
                forward_model.response_structure = response_structure

                # now, we need to translate back to the tensor-world
                jac_torch = th.from_numpy(jac_numpy)
                y_torch = th.from_numpy(response_numpy)

                # the jacobian is recorded, and the model evaluation returned
                ctx.save_for_backward(jac_torch)
                return y_torch

            @staticmethod
            def backward(ctx: Any, *dl_dy_tuple: th.Tensor) -> tuple:
                """
                From the torch.autograd.Function docstring (modified): It must accept a
                context object as the first argument (ctx), followed by as many outputs
                as the 'forward' method of this class returned (None will be passed for
                non-tensor outputs of the 'forward' method), and it should return as
                many tensors, as there were inputs to 'forward' method. Each argument of
                this method (i.e., the 'backward' method) is the gradient w.r.t the
                given output, and each returned value should be the gradient w.r.t. the
                corresponding input. If an input is either not a Tensor or a Tensor not
                requiring gradients, you can just pass None as a gradient for that
                particular input.

                Parameters
                ----------
                ctx
                    This context-object is for derivative-bookkeeping.
                dl_dy_tuple
                    The jacobian of the likelihood with respect to the forward model
                    evaluations, i.e., dl/dy. Since the likelihood returns a scalar
                    value, this is a tuple of length 1.

                Returns
                -------
                    Each element i of the tuple is a tensor that represents the
                    derivative dl/dtheta_i. The order must correspond to the order in
                    the input arguments of the forward-method after the ctx argument.
                    An element is None, if it does not influence the value of the
                    likelihood function.
                """

                # this is dy/dtheta (the gradient with respect to the forward model
                # evaluation) stored from the forward pass; note that the '[0]' at the
                # end is necessary since ctx.saved_tensors is a tuple of all saved
                # tensors; here, there is only one element in it, but we still need to
                # request it
                dy_dtheta = ctx.saved_tensors[0]

                # this is dl/dtheta (the gradient of the likelihood function with
                # respect to the model parameters theta); note that dl_dy_tuple is a
                # tuple of length 1 since the likelihood is a scalar (and not a vector-
                # valued) function
                dl_dy = dl_dy_tuple[0]

                # now dl/dtheta (grad) is computed using the chain rule (which reads
                # dl/dy * dy/dtheta here)
                grad_total = th.matmul(dl_dy, dy_dtheta)

                # the returned gradients must be in tuple form; note that
                # ctx.needs_input_grad is a list with True/False elements of the same
                # lengths as there are parameters in theta; elements in theta that are
                # not being sampled (for example the value of an input sensor of the
                # forward model) are not required to have their gradients evaluated,
                # so these elements will have 'False' entries in ctx.needs_input_grad
                return_val = [None] * self.problem.n_latent_prms
                j = 0
                for i, dim in enumerate(forward_model.input_structure.values()):
                    if ctx.needs_input_grad[i]:
                        return_val[i] = grad_total[j : j + dim]  # type: ignore
                        j += dim  # type: ignore
                    else:
                        j += 1
                return tuple(return_val)

        return self._only_values(Autograd.apply)

    def evaluate_model_response(
        self, theta: th.Tensor, experiment_names: Union[str, List[str], None] = None
    ) -> dict:
        """
        Evaluates the model response for each forward model for the given parameter
        vector theta and the given experiments.

        Parameters
        ----------
        theta
            A numeric vector for which the model responses should be evaluated. Which
            parameters these numbers refer to can be checked by calling self.theta_
            explanation() once the problem is set up.
        experiment_names
            Contains the names of all or some of the experiments added to the inference
            problem. If this argument is None (which is a common use case) then all
            experiments defined in the problem (self.experiments) are used. The names
            provided here define the experiments that the fwd. model is evaluated for.

        Returns
        -------
        model_response_dict
            The first key is the name of the experiment. The values are dicts which
            contain the forward model's output sensor's names as keys have the
            corresponding model responses as values.
        """

        # if experiments is not further specified all experiments added to the problem
        # will be accounted for when computing the model error
        if experiment_names is None:
            experiment_names = [*self.problem.experiments.keys()]
        else:
            # make sure that a given string is converted into a list
            experiment_names = make_list(experiment_names)

        # first, loop over all forward models, and then, over all experiments that are
        # associated with the corresponding model
        model_response_dict = {}
        for fwd_name, fwd_model_wrapped in self.wrapped_forward_models.items():
            forward_model = self.problem.forward_models[fwd_name]
            # get the model parameters for the considered forward model
            prms_model = self.problem.get_parameters(theta, forward_model.prms_def)
            # get all experiments referring to the considered forward model
            relevant_experiment_names = self.problem.get_experiment_names(
                forward_model_names=fwd_name, experiment_names=experiment_names
            )
            # evaluate the forward model for each relevant experiment
            for exp_name in relevant_experiment_names:
                exp_dict = self.problem.experiments[exp_name]
                # prepare the model input values from the experimental data
                sensor_values = exp_dict["sensor_values"]
                exp_inp = {
                    input_sensor.name: sensor_values[input_sensor.name]
                    for input_sensor in forward_model.input_sensors
                }
                inp = {**exp_inp, **prms_model}  # adds the two dictionaries
                # finally, evaluate the forward model for this experiment; note that the
                # additional effort here is necessary, since the wrapped forward model
                # returns a numeric vector that still has to be translated to the
                # dictionary format
                response = fwd_model_wrapped(inp)
                res = forward_model.response_structure
                i = 0
                for key in forward_model.response_structure.keys():
                    n_numbers = forward_model.response_structure[key]
                    res[key] = response[i : i + n_numbers]
                    i += n_numbers
                model_response_dict[exp_name] = res

        return model_response_dict

    def get_theta_samples(self) -> th.Tensor:
        """
        Provides a list of latent-parameter samples in form of torch.Tensors.

        Returns
        -------
            The sampled values based on the latent parameter's priors.
        """
        # even if a list is returned by this function, we initialize a dict here
        # to enable the action in the if-statement in the loop below
        pyro_parameter_samples = dict()  # type: dict
        # we have to sample each parameter for which a prior is defined
        for ref_prm, prior_obj in self.priors.items():
            hyperprms_dict = {}  # type: dict
            for name in prior_obj.hyperparameters.keys():
                if name in pyro_parameter_samples:
                    hyperprms_dict[name] = pyro_parameter_samples[name]
                else:
                    # this is the default case, where the prior's parameters
                    # (i.e. the hyperparameters) are simply constants
                    hyperprms_dict[name] = self.problem.parameters[name].value
            # this is where the parameter's sample is generated with pyro
            sample = prior_obj.sample(hyperprms_dict)  # type: ignore
            if sample.dim() == 0:
                sample = sample.reshape(1)
            pyro_parameter_samples[ref_prm] = sample
        return th.cat(tuple(pyro_parameter_samples.values()))

    def loglike(self, theta: th.Tensor):
        """
        Evaluates the log-likelihood function of the problem at theta.

        Parameters
        ----------
        theta
            A vector of pyro.samples (i.e. tensors) for which the log-likelihood
            function should be evaluated.
        """
        # compute the contribution to the log-likelihood function for each noise model
        # and sum it all up
        for noise_model in self.noise_models:
            # compute the model response for the noise model's experiment_names
            model_response = self.evaluate_model_response(
                theta, noise_model.experiment_names
            )
            # get the tensors for the noise model's parameters
            prms_noise = self.problem.get_parameters(theta, noise_model.prms_def)
            # evaluate the loglike-contribution for the noise model
            noise_model.sample_cond_likelihood(model_response, prms_noise)

    def posterior_model(self):
        """
        Returns the sampled log-likelihood in form of a torch.Tensor.

        Returns
        -------
        ll : torch.Tensor
            The evaluated log-likelihood function for the given theta-vector.
        """
        theta = self.get_theta_samples()
        return self.loglike(theta)

    def run_mcmc(
        self,
        n_walkers: int = 1,
        n_steps: int = 300,
        n_initial_steps: int = 30,
        step_size: float = 0.1,
        **kwargs,
    ) -> az.data.inference_data.InferenceData:
        """
        Runs MCMC with NUTS kernel for the InferenceProblem the PyroSolver was
        initialized with and returns the results as an arviz InferenceData obj.

        Parameters
        ----------
        n_walkers
            Number of walkers used by the estimator.
        step_size
            The step size of the NUTS sampler.
        n_steps
            The number of steps the sampler takes.
        n_initial_steps
            The number of steps for the burn-in phase.
        kwargs
            Additional keyword arguments passed to NUTS.

        Returns
        -------
        inference_data
            Contains the results of the sampling procedure.
        """

        # log which solver is used
        logger.info(
            f"Solving problem using pyro's NUTS sampler with {n_initial_steps} "
            f"+ {n_steps} samples, ..."
        )
        logger.info(f"... {n_walkers} chains and a step size of {step_size:.3f}")
        if kwargs:
            logger.info("Additional NUTS options:")
            print_dict_in_rows(kwargs, printer=logger.info)
        else:
            logger.info("No additional NUTS options specified")

        # prepare the sampling with the requested parameters
        logger.debug("Setting up NUTS sampler")
        th.manual_seed(self.seed)
        kernel = NUTS(self.posterior_model, step_size=step_size, **kwargs)
        logger.debug("Starting sampling (warmup + main)")
        time.sleep(0.1)  # for logging; otherwise no time for new line

        # this is where the actual sampling happens
        start = time.time()
        mcmc = MCMC(
            kernel,
            num_samples=n_steps,
            warmup_steps=n_initial_steps,
            num_chains=n_walkers,
            disable_progbar=not self.show_progress,
        )
        mcmc.run()
        end = time.time()

        # the following modification of the mcmc-object is necessary in cases (that
        # occur so far only when using prior-priors) where the samples of a 1D parameter
        # are saved in a 3D tensor instead of a 2D tensor; note that only a reshape of
        # the data occurs without changing the samples themselves
        for prm_name, samples in mcmc._samples.items():
            if len(samples.shape) > 2 and self.problem.parameters[prm_name].dim == 1:
                mcmc._samples[prm_name] = th.reshape(
                    mcmc._samples[prm_name], samples.shape[:2]
                )

        # log out the results of the process
        runtime_str = pretty_time_delta(end - start)
        logger.info(
            f"Sampling of the posterior distribution completed: "
            f"{n_steps} steps and {n_walkers} chains."
        )
        logger.info(f"Total run-time (including warmup): {runtime_str}.")
        logger.info("")
        logger.info("Summary of sampling results")
        with contextlib.redirect_stdout(stream_to_logger("INFO")):  # type: ignore
            mcmc.summary()
        self.raw_results = mcmc

        # create a summary dictionary similar to the one created by EmceeSolver
        self.summary = {"mean": {}, "median": {}, "sd": {}, "q05": {}, "q95": {}}
        stat_name_map = {
            "mean": "mean",
            "median": "median",
            "std": "sd",
            "5.0%": "q05",
            "95.0%": "q95",
        }
        for prm_name, prm_summary_dict in pyro.infer.mcmc.util.summary(
            mcmc._samples
        ).items():
            for stat, val in prm_summary_dict.items():
                if stat in stat_name_map:
                    if val.numel() == 1:
                        val_conv = float(val)
                    else:
                        # this is for vector parameters
                        val_conv = val.numpy()
                    self.summary[stat_name_map[stat]][prm_name] = val_conv

        # translate the results to a common data structure and return it
        var_names = self.problem.get_theta_names(tex=False, components=False)
        var_names_tex = self.problem.get_theta_names(tex=True, components=False)
        name_dict = {
            var_name: var_name_tex
            for var_name, var_name_tex in zip(var_names, var_names_tex)
        }
        # the following warning-filter is intended to hide the DepreciationWarning that
        # is currently always raised in the arviz.from_pyro method due to using np.bool
        # instead of bool; as soon as this issue is fixed within arviz, the
        # warning-filter can be removed
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            inference_data = az.from_pyro(mcmc, log_likelihood=False)
        inference_data.rename(name_dict, groups="posterior", inplace=True)

        return inference_data



# standard library
from typing import Union, List, Optional, TYPE_CHECKING

# third party imports
import numpy as np
import matplotlib

# local imports
from probeye.subroutines import translate_prms_def

# this is required to prevent a circular import-structure
if TYPE_CHECKING:  # pragma: no cover
    from probeye.definition.parameter import Parameters


class PriorBase:
    """
    Template class for prior definitions. Note that the main motivation of how this
    class is implemented was to avoid storing any numeric values for any of the priors
    parameters within the prior object.
    """

    def __init__(
        self,
        ref_prm: str,
        prms_def: Union[str, dict, List[Union[str, dict]]],
        name: str,
        prior_type: str,
    ):
        """
        Parameters
        ----------
        ref_prm
            The name of the latent parameter the prior refers to.
        prms_def
            A list of strings, or list of one-element-dicts defining the prior's
            parameter names. E.g. ['loc_a', 'scale_a'] or {'loc_a': 'loc_a', 'scale_a':
            'scale_a'}. The latter example is the notation for the use of global and
            local names, which should not be necessary for the definition of prior-
            parameters. A special case is the uninformative prior (see below) which
            hasn't got an parameters. So, in this case prms_def might also be None.
        name
            Defining the priors name.
        prior_type
            Stating the prior's type, e.g. "normal distribution". This is just used for
            printing information on the prior.
        """
        # write arguments to attributes
        self.ref_prm = ref_prm
        self.name = name
        self.prior_type = prior_type

        # these dictionaries contain the <global name>:<local name> pairs of the prior's
        # parameters with and without the reference variable; if for example you define
        # a normal prior for parameter 'a' with location, then 'a' is the reference
        # variable; note that the conversion to a dictionary via list2dict is due to the
        # possibility of using local parameter names, which however is not intended to
        # be used for priors
        self.hyperparameters, _ = translate_prms_def(prms_def)
        self.prms_def, _ = translate_prms_def(
            [ref_prm] + [*self.hyperparameters.values()]
        )

    def __str__(self) -> str:
        """
        Allows printing an object of this class.

        Returns
        -------
        s
            A string containing details on the respective prior.
        """
        s = f"{self.prior_type} for '{self.ref_prm}', prms={self.prms_def}"
        return s

    def plot(
        self,
        ax: "matplotlib.axes",
        prms: "Parameters",
        x: Optional[np.ndarray] = None,
        n_points: int = 200,
        n_sigma: Union[int, float] = 2,
        color: str = "darkorange",
        rotate: bool = False,
    ):
        """
        Plots the prior-pdf to a given axis object.

        Parameters
        ----------
        ax
            The axis object to plot the prior-pdf on.
        prms
            The parameters of the problem at hand.
        x
            The points where the prior-pdf should be evaluated at. If None is
            given, x will be derived from the x-limits of the given ax-object.
        n_points
            The number of points of the prior-pdf graph. Only effective when
            x is None.
        n_sigma
            Defines the x-range of a normal prior via mean plus/minus n_sigma
            times the standard deviation.
        color
            The line-color of the prior-pdf's graph.
        rotate
            If True, the x- and y-axis are switched.
        """

        if self.prior_type == "normal":
            mu = prms[self.prms_def[f"loc_{self.ref_prm}"]].value
            sigma = prms[self.prms_def[f"scale_{self.ref_prm}"]].value
            # proceed, only if both values are constants and not latent
            # parameters themselves
            if mu and sigma:
                if x is None:
                    left = mu - n_sigma * sigma
                    right = mu + n_sigma * sigma
                    x = np.linspace(left, right, n_points)
                else:
                    left = min(x)
                    right = max(x)
                y = 1 / (sigma * np.sqrt(2 * np.pi))
                y *= np.exp(-0.5 * ((x - mu) / sigma) ** 2)
                margin = 0.025 * abs(right - left)
                if rotate:
                    ax.plot(y, x, label="prior", color=color)
                    ax.set_ylim((left - margin, right + margin))
                else:
                    ax.plot(x, y, label="prior", color=color)
                    ax.set_xlim((left - margin, right + margin))

        elif self.prior_type == "uniform":
            a = prms[self.prms_def[f"low_{self.ref_prm}"]].value
            b = prms[self.prms_def[f"high_{self.ref_prm}"]].value
            # proceed, only if both values are constants and not latent
            # parameters themselves
            if a and b:
                y = np.zeros(n_points)
                y[1:-1] = np.ones(n_points - 2) / (b - a)
                if x is None:
                    x = np.linspace(a, b, n_points)
                    y[0], y[-1] = 0, 0
                else:
                    y[0] = 0 if (x[0] <= a) else y[1]
                    y[-1] = 0 if (x[-1] >= b) else y[-2]
                margin = 0.025 * abs(b - a)
                if rotate:
                    ax.plot(y, x, label="prior", color=color)
                    ax.set_ylim((a - margin, b + margin))
                else:
                    ax.plot(x, y, label="prior", color=color)
                    ax.set_xlim((a - margin, b + margin))



# standard library imports
from typing import Union, List, Optional, TYPE_CHECKING


# third party imports
import torch as th
import pyro
import pyro.distributions as dist

# local imports
from probeye.definition.noise_model import NormalNoiseModel

# imports only needed for type hints
if TYPE_CHECKING:  # pragma: no cover
    from probeye.definition.sensor import Sensor


class NormalNoise(NormalNoiseModel):
    """
    A general Gaussian (normal) noise model with or without correlations. Note
    that currently, there are no correlation capabilities defined.
    """

    def __init__(
        self,
        target_sensor: "Sensor",
        prms_def: Union[str, List[Union[str, dict]], dict],
        name: Optional[str] = None,
        corr: Optional[str] = None,
        corr_model: Optional[str] = "exp",
        noise_type: str = "additive",
    ):
        """
        For information on most of the above arguments check out the docstring
        of the parent class's __init__ method.
        """
        # initialize the super-class (NormalNoiseModel) based on the given input
        super().__init__(
            prms_def=prms_def,
            sensors=target_sensor,
            name=name,
            corr=corr,
            corr_model=corr_model,
            noise_type=noise_type,
        )

        # the following attributes are not considered in the parent class
        self.target_sensor = target_sensor

    def error(self, model_response_dict: dict) -> dict:
        """
        Computes the model error for all of the noise model's experiments and returns
        them in a dictionary that is sorted by output sensor_values.

        Parameters
        ----------
        model_response_dict
            The first key is the name of the experiment. The values are dicts which
            contain the forward model's output sensor's names as keys have the
            corresponding model responses as values.

        Returns
        -------
        model_error
            A dictionary with the keys being the noise model's sensor names, and torch.
            Tensors representing the model errors as values.
        """

        # prepare the dictionary keys
        model_error_dict = {name: th.Tensor([]) for name in self.sensor_names}

        # fill the dictionary with model error vectors
        for exp_name in self.experiment_names:
            exp_dict = self.problem_experiments[exp_name]
            ym_dict = model_response_dict[exp_name]
            ye_dict = exp_dict["sensor_values"]
            me_dict = self.error_function(ym_dict, ye_dict)
            model_error_dict = {
                name: th.cat((model_error_dict[name], me_dict[name]))
                for name in self.sensor_names
            }
        return model_error_dict

    def sample_cond_likelihood(self, model_response: dict, prms: dict) -> pyro.sample:
        """
        Creates a likelihood-sample conditioned on the observed errors.

        model_response_dict
            The first key is the name of the experiment. The values are dicts which
            contain the forward model's output sensor's names as keys have the
            corresponding model responses as values.
        prms
            Contains parameter name:value pairs.
        """
        std = prms["std"]
        mean = 0.0 if self.zero_mean else prms["mean"]
        # compute the model error; note that this mode has exactly one sensor
        model_error_vector = self.error(model_response)[self.target_sensor.name]
        pyro.sample(f"lkl_{self.name}", dist.Normal(mean, std), obs=model_error_vector)


def translate_noise_model(noise_base: NormalNoiseModel) -> NormalNoise:
    """
    Translates a given instance of NoiseBase (which is essentially just a description
    of the noise model without computing-methods) to a specific noise model object which
    does contain computing-methods (e.g. compute the log-likelihood contribution).

    Parameters
    ----------
    noise_base
        An instance of NoiseBase which contains basic information on the noise
        model but no computing-methods.

    Returns
    -------
    noise_object
        An instance of a specific noise model class with computing capabilities.
        Examples for such classes are given above in this file.
    """

    # this is the noise classes currently defined (see code above)
    noise_classes = {"normal": NormalNoise}

    # this is where the translation happens
    noise_object = noise_classes[noise_base.dist](
        target_sensor=noise_base.sensors[0],
        prms_def=noise_base.prms_def,
        name=noise_base.name,
        corr=noise_base.corr,
        corr_model=noise_base.corr_model,
        noise_type=noise_base.noise_type,
    )

    # here, we take the assigned experiments from the base object
    noise_object.experiment_names = noise_base.experiment_names
    noise_object.problem_experiments = noise_base.problem_experiments

    return noise_object



# standard library
from typing import Union, List, Optional

# third party imports
import numpy as np

# local imports
from probeye.definition.sensor import Sensor
from probeye.subroutines import make_list, translate_prms_def


class NoiseModelBase:
    def __init__(
        self,
        dist: str,
        prms_def: Union[str, List[Union[str, dict]], dict],
        sensors: Union[Sensor, List[Sensor]],
        name: Optional[str] = None,
        corr: Optional[str] = None,
        corr_model: Optional[str] = None,
        noise_type: str = "additive",
    ):
        """
        Parameters
        ----------
        dist
            A string specifying the probability distribution the noise model is based
            on, e.g. 'normal'.
        prms_def
            A list of parameter names (strings) defining how a noise parameter vector
            given to the loglike_contribution method is interpreted. For example:
            prms_def = ['mu', 'sigma'] means that the noise parameter vector has two
            elements, the first of which gives the value of 'mu' and the second gives
            the value of 'sigma'.
        sensors
            Sensor objects that are required to evaluate the noise model.
        name
            Unique name of the noise model. This name is None, if the user does not
            specify it when adding the noise model to the problem. It is then named
            automatically before starting the inference engine.
        corr
            Defines the correlation model. So far this is just a placeholder. It is not
            clear yet how exactly the correlation should be defined. When it is set to
            None, all sensors/sensor elements are independent.
        corr_model
            Defines the correlation function to be used in case corr isn't None.
        noise_type
            Either 'additive', 'multiplicative' or 'other'. Defines if the error is
            computed via [prediction - measurement] ('additive') or via [prediction/
            measurement-1] ('multiplicative') or in some 'other' i.e.,
            non-standard fashion.
        """
        self.dist = dist
        self.prms_def, self.prms_dim = translate_prms_def(prms_def)
        self.sensors = make_list(sensors)
        self.sensor_names = [sensor.name for sensor in self.sensors]
        self.name = name
        self.corr = corr
        self.corr_model = corr_model
        self.noise_type = noise_type
        # this is a list of experiment names, that relate to the noise model; the list
        # will be filled after experiments have been added to the InferenceProblem and
        # the problem definition is complete; in this case call InferenceProblem.
        # assign_experiments_to_noise_models() to fill it with the corresponding names
        self.experiment_names = []  # type: List[str]
        # as soon as defined, this attribute will be a pointer to the inference
        # problems experiments (it will be used for consistency checks)
        self.problem_experiments = {}  # type: dict

        # set the error_function depending on the noise-type
        if noise_type == "additive":
            self.error_function = self.error_function_additive
        elif noise_type == "multiplicative":
            self.error_function = self.error_function_multiplicative
        elif noise_type == "other":
            self.error_function = self.error_function_other
        else:
            raise ValueError(
                f"Encountered unknown noise_type: '{noise_type}'. The noise_type must "
                f"be either 'additive', 'multiplicative' or 'other'."
            )

    def add_experiments(self, experiment_names_: Union[str, List[str]]):
        """
        Adds experiment names to the noise model. When the noise model is evaluated it
        will only be evaluated for those experiments added here.

        Parameters
        ----------
        experiment_names_
            Names (strings) of experiments from the InferenceProblem that should be
            added to the noise model.
        """
        # check if the given experiments are compatible with the noise model with
        # respect to the sensors
        experiment_names = make_list(experiment_names_)
        forward_models = set()
        for exp_name in experiment_names:
            exp_dict = self.problem_experiments[exp_name]
            forward_models.add(exp_dict["forward_model"])
            sensor_names_exp = [*exp_dict["sensor_values"].keys()]
            for sensor_name in self.sensor_names:
                if sensor_name not in sensor_names_exp:
                    raise RuntimeError(
                        f"Experiment '{exp_name}' does not contain a sensor "
                        f"'{sensor_name}' which is required for the evaluation of the "
                        f"noise model."
                    )
        # check if the given experiments all refer to one forward model
        if len(forward_models) > 1:
            raise RuntimeError(
                f"The given experiments refer to more than one forward model!"
            )
        # check if one of the given experiments have been added before
        for exp_name in experiment_names:
            if exp_name in self.experiment_names:
                raise RuntimeError(
                    f"The experiment '{exp_name}' has already been added to this noise "
                    f"model. Something might be wrong here."
                )
        self.experiment_names += experiment_names

    def error(self, model_response_dict: dict) -> dict:
        """
        Computes the model error for all of the noise model's experiments and returns
        them in a dictionary that is sorted by output sensor_values.

        Parameters
        ----------
        model_response_dict
            The first key is the name of the experiment. The values are dicts which
            contain the forward model's output sensor's names as keys have the
            corresponding model responses as values.

        Returns
        -------
        model_error
            A dictionary with the keys being the noise model's sensor names, and 1D
            numpy arrays representing the model errors as values.
        """
        # prepare the dictionary keys
        model_error_dict = {name: np.array([]) for name in self.sensor_names}

        # fill the dictionary with model error vectors
        for exp_name in self.experiment_names:
            exp_dict = self.problem_experiments[exp_name]
            ym_dict = model_response_dict[exp_name]
            ye_dict = exp_dict["sensor_values"]
            me_dict = self.error_function(ym_dict, ye_dict)
            model_error_dict = {
                name: np.append(model_error_dict[name], me_dict[name])
                for name in self.sensor_names
            }

        return model_error_dict

    def error_function_additive(self, ym_dict: dict, ye_dict: dict) -> dict:
        """
        Evaluates the additive model error for each of the noise model' sensors.

        Parameters
        ----------
        ym_dict
            The computed values for the model's output sensor_values.
        ye_dict
            The measured values for the model's output sensor_values.

        Returns
        -------
        error_dict
            The computed model error for the model's output sensor_values.
        """
        # for each sensor, its own error metric is used to compute the error
        error_dict = {name: ym_dict[name] - ye_dict[name] for name in self.sensor_names}
        return error_dict

    def error_function_multiplicative(self, ym_dict: dict, ye_dict: dict) -> dict:
        """
        Evaluates the multiplicative model error for each of the noise model's sensors.

        Parameters
        ----------
        ym_dict
            The computed values for the model's output sensor_values.
        ye_dict
            The measured values for the model's output sensor_values.

        Returns
        -------
        error_dict
            The computed model error for the model's output sensor_values.
        """
        # for each sensor, its own error metric is used to compute the error
        error_dict = {
            name: ym_dict[name] / ye_dict[name] - 1.0 for name in self.sensor_names
        }
        return error_dict

    def error_function_other(self, ym_dict: dict, ye_dict: dict) -> dict:
        """
        Non-standard error function self.error_function will point to when self.
        noise_type is set to 'other'. See self.error_function for more information.
        """
        raise NotImplementedError(
            "Your model does not have an non-standard error_function-method yet. If "
            "you want to use it, you need to implement it first."
        )

    def loglike_contribution(
        self, model_response_dict: dict, prms: dict, worst_value: float = -np.infty
    ) -> float:
        """
        Evaluates the log-likelihood function for the given model error and the given
        noise parameter vector. This method has to be overwritten.

        Parameters
        ----------
        model_response_dict
            The first key is the name of the experiment. The values are dicts which
            contain the forward model's output sensor's names as keys have the
            corresponding model responses as values.
        prms
            Dictionary containing parameter name:value pairs.
        worst_value
            This value is returned when this method does not result in a numeric value.
            This might happen for example when the given parameters are not valid (for
            example in case of a negative standard deviation). The returned value in
            such cases should represent the worst possible value of the contribution.

        Returns
        -------
        ll
            The evaluated log-likelihood function.
        """
        raise NotImplementedError(
            "Your model does not have a loglike_contribution-method. You need to "
            "define this method so you can evaluate your noise model."
        )


class NormalNoiseModel(NoiseModelBase):
    """
    A general Gaussian (normal) noise model with or without correlations.
    """

    def __init__(
        self,
        prms_def: Union[str, List[Union[str, dict]], dict],
        sensors: Union[Sensor, List[Sensor]],
        name: Optional[str] = None,
        corr: Optional[str] = None,
        corr_model: Optional[str] = None,
        noise_type: str = "additive",
    ):
        """
        See docstring of NoiseModelBase for information on the input arguments.
        """

        # initialize the base class with given input
        super().__init__(
            "normal",
            prms_def,
            sensors,
            name=name,
            corr=corr,
            corr_model=corr_model,
            noise_type=noise_type,
        )

        # check that at the standard deviation is provided (this can be either as a
        # constant or a latent parameter, but it has to be given)
        if "std" not in [*self.prms_def.values()]:
            raise RuntimeError(
                "The standard deviation 'std' was not provided in prms_def!"
            )

        # the mean value(s) do not have to be stated explicitly; if they are not given,
        # the are assumed to be zero
        self.zero_mean = True
        if "mean" in [*self.prms_def.values()]:
            self.zero_mean = False



# standard library
from typing import Union, List, Dict
import copy as cp
import numpy as np

# local imports
from probeye.definition.sensor import Sensor
from probeye.subroutines import translate_prms_def
from probeye.subroutines import make_list
from probeye.subroutines import len_or_one


class ForwardModelBase:
    """
    This class serves as a base class for any forward model. When you want to define a
    specific forward model, you need to derive your own class from this one, and then
    define the '__call__' method. The latter essentially describes the model function
    mapping the model input to the output.
    """

    def __init__(
        self,
        prms_def_: Union[str, List[Union[str, dict]], dict],
        input_sensors: Union[Sensor, List[Sensor]],
        output_sensors: Union[Sensor, List[Sensor]],
    ):
        """
        Parameters
        ----------
        prms_def_
            Contains the model's latent parameter names. The list may only contain
            strings or one-element dictionaries. It could look, for example, like
            [{'a': 'm'}, 'b']. The one-element dictionaries account for the possibility
            to define a local name for a latent parameter that is different from the
            global name. In the example above, the latent parameter with the global name
            'a' will be referred to as 'm' within the model. So, the one-element dicts
            have the meaning {<global name>: <local name>}. String-elements are
            interpreted as having similar local and global names. Note that the local-
            name option will not be required most of the times. The input from global to
            local name can also be provided as a dict. In the example above it would
            look like {'a': 'm', 'b': 'b'}.
        input_sensors
            Contains sensor-objects structuring the model input.
        output_sensors
            Contains sensor-objects structuring the model output.
        """

        # convert the given parameter names to a dictionary with global names as keys
        # and local names as values
        self.prms_def, self.prms_dim = translate_prms_def(prms_def_)

        # other attributes
        self.input_sensors = make_list(input_sensors)
        self.output_sensors = make_list(output_sensors)

        # this attribute might be used to write the forward model's input structure to;
        # it has the same structure like the 'inp' argument of the response method, but
        # instead of the input channel's values it states the input channels number of
        # elements
        self.input_structure = {ic: 0 for ic in self.input_channel_names}

        # this attributes might be used by inference engines that need a forward model
        # wrapper, which only returns numeric vectors; for reconstructing the response
        # dictionary from the numeric vector, one needs to know the response
        # dictionary's structure; this dictionaries will then contain the same keys as
        # the response method's return dictionary, while the values will be the number
        # of elements contained in the values; e.g., {'x': np.array([0, 0.1, 0.2]),
        # 'a': 3.7} will have a structure  of {'x': 3, 'a': 1}; this attr. is not used
        # by all inference engines
        self.response_structure = {os.name: 0 for os in self.output_sensors}

    @property
    def input_sensor_names(self) -> List[str]:
        """Provides input_sensor_names attribute."""
        return [sensor.name for sensor in self.input_sensors]

    @property
    def input_channel_names(self) -> List[str]:
        """Provides input_channel_names attribute."""
        return self.input_sensor_names + [*self.prms_def.values()]

    @property
    def output_sensor_names(self) -> List[str]:
        """Provides input_sensor_names attribute."""
        return [sensor.name for sensor in self.output_sensors]

    def response(self, inp: dict) -> dict:
        """
        Evaluates the model response and provides computed results for all of the
        model's output sensors. This method must be overwritten by the user.

        Parameters
        ----------
        inp
            Contains both the exp. input data and the  model's parameters. The keys are
            the names, and the values are their numeric values.

        Returns
        -------
        dict
            Contains the model response (value) for each output sensor, referenced by
            the output sensor's name (key).
        """
        raise NotImplementedError(
            "Your model does not have a proper 'response'-method yet. You need  to "
            "define this method, so you can evaluate your model."
        )

    def __call__(self, inp: dict) -> dict:
        """
        Calls the self.response method. Shortens internal forward model calls.
        """
        return self.response(inp)

    def jacobian(self, inp: dict) -> Dict[str, dict]:
        """
        Numerically computes the Jacobian matrix of the forward model and returns it in
        form of a dictionary. Note that this method should be overwritten, if there is a
        more efficient way to compute the jacobian, for example, when one can compute
        the Jacobian analytically.

        Parameters
        ----------
        inp
            Contains both the exp. input data and the  model's parameters. The keys are
            the names, and the values are their numeric values.

        Returns
        -------
        jac_dict
            The Jacobian matrix in dict-form: The keys are the names of the forward
            model's output sensors. The values are dictionaries with the forward model's
            input channel's names as keys and the derivatives or Nones as values.
            Derivatives are only provided for the model's parameters, see self.prms_def.
            For all other input channels (e.g., measurements from an experiment) None is
            written to the dictionary's values, since they are not required by sampling
            routines. To give an example: the element jac['y']['a'] would give the
            derivative dy/da, and jac['y'] would give the gradient of the fwd. model's
            y-computation with respect to the input channels in a dictionary-format.
        """
        # eps is the machine precision; it is needed to compute the step size of the
        # central difference scheme below; note that this refers to single precision
        # (float32) since the processed arrays might be defined in float32, in which
        # case using the eps of double precision (float64) would not work since the
        # step size would be too small
        eps = np.finfo(np.float32).eps
        # the following evaluations are needed in the for-loop; they are put here so
        # they are not repeatedly evaluated (to the same value) during the for-loop
        sqrt_eps = np.sqrt(eps)
        response_dict_center = self.response(inp)
        inp_right = cp.deepcopy(inp)  # required to prevent side effects
        # prepare the dictionary; this structure needs to be external from the main loop
        # below since the filling of the dictionary could only be efficiently done in
        # the format jac_dict[prm_name][os_name] which is less readable; the format
        # created in the implemented way is easier to to read since jac['y']['a']
        # corresponds to dy/da in jac['y'] is the gradient of y with respect to theta
        jac_dict = {}  # type: Dict[str, dict]
        for output_sensor in self.output_sensors:
            jac_dict[output_sensor.name] = {}
            for prm_name in inp.keys():
                if prm_name in self.prms_def.values():  # values are local names
                    nrows = len_or_one(response_dict_center[output_sensor.name])
                    ncols = len_or_one(inp[prm_name])
                    value = np.zeros((nrows, ncols))  # type: Union[np.ndarray, None]
                else:
                    value = None
                jac_dict[output_sensor.name][prm_name] = value
        for prm_name, prm_value in inp.items():
            # derivatives only need to be computed for the model's parameters, not for
            # the input that comes from the experiments; e.g., in case of y=mx+b with
            # parameters m, b, one does not need dy/dx
            if prm_name not in self.prms_def.values():  # values are local names
                continue
            # the following loop accounts for the fact, that parameters can be
            # multidimensional
            for i, x in enumerate(np.atleast_1d(inp[prm_name])):
                # the following formula for the step size is NOT taken from the
                # literature; in the literature, a common recommended choice for the
                # step size h given x is not 0 is h = sqrt_eps * x, see for example:
                # https://en.wikipedia.org/wiki/Numerical_differentiation; we added the
                # term '+ sqrt_eps' below to also cover the cases where x actually is
                # zero (or very close to 0)
                h = sqrt_eps * x + sqrt_eps
                inp_right[prm_name] = x + h
                response_dict_right = self.response(inp_right)
                for output_sensor in self.output_sensors:
                    # the simple forward scheme should be sufficient for most
                    # applications since the Jacobian will only be used as info for
                    # choosing the next sample; for that purpose it is secondary if it
                    # contains small numerical errors
                    jac_dict[output_sensor.name][prm_name][:, i] = (
                        response_dict_right[output_sensor.name]
                        - response_dict_center[output_sensor.name]
                    ) / h
                inp_right[prm_name] = inp[prm_name]  # resetting perturbed value
        return jac_dict

    def jacobian_dict_to_array(
        self,
        inp: dict,
        jac_dict: dict,
        n_inp_dim: int,
    ) -> np.ndarray:
        """
        Converts the Jacobian in dict-format (computed by the above 'jacobian' method)
        into a numpy array. This method is external to the above 'jacobian' method, so
        that it is easier for a user to overwrite the it (i.e., the 'jacobian' method)
        without also having to define the conversion into an array.

        Parameters
        ----------
        inp
            See docstring of the 'jacobian'-method above.
        jac_dict
            See docstring of the 'jacobian'-method above.
        n_inp_dim
            The added-up dimensions of the forward model's input channels, i.e., of all
            model parameters and other input variables.

        Returns
        -------
        jac
            Similar structure as the conventional Jacobi matrix with respect to the
            columns and rows (i.e. the rows are the different gradients and the columns
            are derivatives with respect to one fixed parameter).
        """

        # n1 is the number of the forward model's output sensors; n2 is the dimension of
        # the forward model's input channels, i.e., the added up dimensions of the input
        # sensors and the number of the forward model's parameters; finally, n3 is the
        # maximum number of elements in the n2 input channels; the model's parameters
        # are usually scalars, but the input sensors might be vectors with more than one
        # element
        n1 = len(self.output_sensors)
        n2 = n_inp_dim
        n3 = max([len_or_one(v) for v in inp.values()])
        jac = np.zeros((n1 * n3, n2))
        for i, prm_dict in enumerate(jac_dict.values()):
            idx_start = i * n3
            j = 0
            for derivative in prm_dict.values():
                if derivative is None:
                    # in this case, the input variable is not a parameter; the
                    # corresponding column in the Jacobian will remain zero, which is
                    # ok, since it won't be used to compute anything
                    j += 1
                    continue
                nvals = derivative.shape[0]
                ncomp = derivative.shape[1]
                idx_end = idx_start + nvals
                jac[idx_start:idx_end, j : (j + ncomp)] = derivative
                j += ncomp
        return jac



# standard library imports
import unittest

# local imports
from probeye.definition.inference_problem import InferenceProblem
from probeye.postprocessing.sampling import create_pair_plot
from probeye.postprocessing.sampling import create_posterior_plot
from probeye.postprocessing.sampling import create_trace_plot


class TestProblem(unittest.TestCase):
    def test_not_implemented_pair_plot(self):
        # check for currently not implemented plotting frameworks
        p = InferenceProblem("TestProblem")
        p.add_parameter("a", "model", prior=("normal", {"loc": 0, "scale": 1}))
        # only one parameter given (no error, just warning)
        create_pair_plot(None, p, plot_with="arviz")
        # add another parameter to have a valid setup
        p.add_parameter("b", "model", prior=("normal", {"loc": 0, "scale": 1}))
        with self.assertRaises(NotImplementedError):
            create_pair_plot(None, p, plot_with="seaborn")
        with self.assertRaises(NotImplementedError):
            create_pair_plot(None, p, plot_with="matplotlib")
        with self.assertRaises(RuntimeError):
            create_pair_plot(None, p, plot_with="other package")

    def test_not_implemented_create_posterior_plot(self):
        # check for currently not implemented plotting frameworks
        p = InferenceProblem("TestProblem")
        with self.assertRaises(NotImplementedError):
            create_posterior_plot(None, p, plot_with="seaborn")
        with self.assertRaises(NotImplementedError):
            create_posterior_plot(None, p, plot_with="matplotlib")
        with self.assertRaises(RuntimeError):
            create_posterior_plot(None, p, plot_with="other package")

    def test_not_implemented_create_trace_plot(self):
        # check for currently not implemented plotting frameworks
        p = InferenceProblem("TestProblem")
        with self.assertRaises(NotImplementedError):
            create_trace_plot(None, p, plot_with="seaborn")
        with self.assertRaises(NotImplementedError):
            create_trace_plot(None, p, plot_with="matplotlib")
        with self.assertRaises(RuntimeError):
            create_trace_plot(None, p, plot_with="other package")


if __name__ == "__main__":
    unittest.main()
